from __future__ import annotations

import os
import time
import torch
import random
import datetime
import threading

from typing import (
    Type,
    Union,
    Any,
    Optional,
    List,
    Tuple,
    Dict,
    Literal,
    Callable,
    TYPE_CHECKING,
    cast,
)
from hashlib import md5

from pibble.api.configuration import APIConfiguration
from pibble.api.exceptions import ConfigurationError
from pibble.util.files import load_json, dump_json

from enfugue.util import logger, check_download, check_make_directory
from enfugue.diffusion.constants import (
    DEFAULT_MODEL,
    DEFAULT_INPAINTING_MODEL,
    CONTROLNET_CANNY,
    CONTROLNET_MLSD,
    CONTROLNET_HED,
    CONTROLNET_SCRIBBLE,
    CONTROLNET_TILE,
    CONTROLNET_INPAINT,
)

__all__ = ["DiffusionPipelineManager"]

if TYPE_CHECKING:
    from diffusers.models import ControlNetModel
    from enfugue.diffusion.upscale import Upscaler
    from enfugue.diffusion.pipeline import EnfugueStableDiffusionPipeline
    from enfugue.diffusion.edge.detect import EdgeDetector


class KeepaliveThread(threading.Thread):
    """
    Calls the keepalive function every <n> seconds.
    """

    INTERVAL = 0.5
    KEEPALIVE_INTERVAL = 10

    def __init__(self, manager: DiffusionPipelineManager) -> None:
        super(KeepaliveThread, self).__init__()
        self.manager = manager
        self.stop_event = threading.Event()

    @property
    def stopped(self) -> bool:
        """
        Returns true IFF the stop event is set.
        """
        return self.stop_event.is_set()

    def stop(self) -> None:
        """
        Stops the thread.
        """
        self.stop_event.set()

    def run(self) -> None:
        """
        The threading run loop.
        """
        last_keepalive = datetime.datetime.now()
        while not self.stopped:
            time.sleep(self.INTERVAL)
            now = datetime.datetime.now()
            if (now - last_keepalive).total_seconds() > self.KEEPALIVE_INTERVAL:
                self.manager.keepalive_callback()
                last_keepalive = now


class DiffusionPipelineManager:
    TENSORRT_STAGES = [
        "unet"
    ]  # TODO: Get others to work with multidiff (clip works but isnt worth it right now)
    TENSORRT_ALWAYS_USE_CONTROLLED_UNET = False  # TODO: Figure out if this is possible

    PIPELINE_CLASS = "enfugue.diffusion.pipeline.EnfugueStableDiffusionPipeline"
    TRT_PIPELINE_CLASS = "enfugue.diffusion.rt.pipeline.EnfugueTensorRTStableDiffusionPipeline"

    DEFAULT_CHUNK = 64
    DEFAULT_SIZE = 512

    _keepalive_thread: KeepaliveThread

    def __init__(self, configuration: Optional[APIConfiguration] = None) -> None:
        self.configuration = APIConfiguration()
        if configuration:
            self.configuration = configuration

    @property
    def safe(self) -> bool:
        """
        Returns true if safety checking should be enabled.
        """
        if not hasattr(self, "_safe"):
            self._safe = self.configuration.get("enfugue.safe", True)
        return self._safe

    @safe.setter
    def safe(self, val: bool) -> None:
        """
        Sets a new value for safety checking. Destroys the pipeline.
        """
        if val != getattr(self, "_safe", None):
            self._safe = val
            del self.pipeline

    @property
    def device(self) -> str:
        """
        Gets the device that will be executed on
        """
        return "cuda" if torch.cuda.is_available() else "cpu"

    @property
    def seed(self) -> int:
        """
        Gets the seed. If there is none, creates a random one once.
        """
        if not hasattr(self, "_seed"):
            self._seed = self.configuration.get("enfugue.seed", random.randint(0, 2**63 - 1))
        return self._seed

    @seed.setter
    def seed(self, new_seed: int) -> None:
        """
        Re-seeds the pipeline. This deletes the generator so it gets re-initialized.
        """
        self._seed = new_seed
        del self.generator

    @property
    def keepalive_callback(self) -> Callable[[], None]:
        """
        A callback function to call during long operations.
        """
        if not hasattr(self, "_keepalive_callback"):
            return lambda: None
        return self._keepalive_callback

    @keepalive_callback.setter
    def keepalive_callback(self, new_callback: Callable[[], None]) -> None:
        """
        Sets the callback
        """
        self._keepalive_callback = new_callback

    @keepalive_callback.deleter
    def keepalive_callback(self) -> None:
        """
        Removes the callback.
        """
        if hasattr(self, "_keepalive_callback"):
            del self._keepalive_callback

    def start_keepalive(self) -> None:
        """
        Starts a thread which will call the keepalive callback every <n> seconds.
        """
        if not hasattr(self, "_keepalive_thread") or not self._keepalive_thread.is_alive():
            keepalive_callback = self.keepalive_callback
            self._keepalive_thread = KeepaliveThread(self)
            self._keepalive_thread.start()

    def stop_keepalive(self) -> None:
        """
        Stops the thread which calls the keepalive.
        """
        if hasattr(self, "_keepalive_thread"):
            if self._keepalive_thread.is_alive():
                self._keepalive_thread.stop()
                self._keepalive_thread.join()
            del self._keepalive_thread

    @property
    def generator(self) -> torch.Generator:
        """
        Creates the generator once, otherwise returns it.
        """
        if not hasattr(self, "_generator"):
            self._generator = torch.Generator(device=self.device)
            self._generator.manual_seed(self.seed)
        return self._generator

    @generator.deleter
    def generator(self) -> None:
        """
        Removes an existing generator.
        """
        if hasattr(self, "_generator"):
            delattr(self, "_generator")

    @property
    def size(self) -> int:
        """
        Gets the base engine size in pixels when chunking (default always.)
        """
        if not hasattr(self, "_size"):
            self._size = int(
                self.configuration.get("enfugue.size", DiffusionPipelineManager.DEFAULT_SIZE)
            )
        return self._size

    @size.setter
    def size(self, new_size: int) -> None:
        """
        Sets the base engine size in pixels.
        """
        if hasattr(self, "_size") and self._size != new_size:
            del self.pipeline
        self._size = new_size

    @property
    def chunking_size(self) -> int:
        """
        Gets the chunking size in pixels.
        """
        if not hasattr(self, "_chunking_size"):
            self._chunking_size = int(
                self.configuration.get("enfugue.chunk.size", DiffusionPipelineManager.DEFAULT_CHUNK)
            )
        return self._chunking_size

    @chunking_size.setter
    def chunking_size(self, new_chunking_size: int) -> None:
        """
        Sets the new chunking size. This doesn't require a restart.
        """
        self._chunking_size = new_chunking_size

    @property
    def chunking_blur(self) -> int:
        """
        Gets the chunking blur in pixels.
        """
        if not hasattr(self, "_chunking_blur"):
            self._chunking_blur = int(
                self.configuration.get("enfugue.chunk.blur", DiffusionPipelineManager.DEFAULT_CHUNK)
            )
        return self._chunking_blur

    @chunking_blur.setter
    def chunking_blur(self, new_chunking_blur: int) -> None:
        """
        Sets the new chunking blur. This doesn't require a restart.
        """
        self._chunking_blur = new_chunking_blur

    @property
    def engine_root(self) -> str:
        """
        Gets the root of the engine.
        """
        root = self.configuration.get("enfugue.engine.root", "~/.cache/enfugue")
        if root.startswith("~"):
            root = os.path.expanduser(root)
        root = os.path.realpath(root)
        check_make_directory(root)
        return root

    @property
    def diffusers_cache_dir(self) -> str:
        """
        Gets the cache for diffusers-downloaded configuration files, base models, etc.
        """
        path = os.path.join(self.engine_root, "cache")
        check_make_directory(path)
        return path

    @property
    def engine_checkpoints_dir(self) -> str:
        """
        Gets where checkpoints are downloaded in.
        """
        path = os.path.join(self.engine_root, "checkpoint")
        check_make_directory(path)
        return path

    @property
    def engine_other_dir(self) -> str:
        """
        Gets where any other weights are download in
        """
        path = os.path.join(self.engine_root, "other")
        check_make_directory(path)
        return path

    @property
    def engine_lora_dir(self) -> str:
        """
        Gets where lora are downloaded in.
        """
        path = os.path.join(self.engine_root, "lora")
        check_make_directory(path)
        return path

    @property
    def engine_inversions_dir(self) -> str:
        """
        Gets where inversions are downloaded in.
        """
        path = os.path.join(self.engine_root, "inversions")
        check_make_directory(path)
        return path

    @property
    def engine_models_dir(self) -> str:
        """
        Gets where models are pushed to after converting to diffusers.
        """
        path = os.path.join(self.engine_root, "models")
        check_make_directory(path)
        return path

    @property
    def model_dir(self) -> str:
        """
        Gets where the current model should be stored.
        """
        model = self.model
        path = os.path.join(self.engine_models_dir, self.model)
        check_make_directory(path)
        return path

    @property
    def model_tensorrt_dir(self) -> str:
        """
        Gets where tensorrt engines will be built per model.
        """
        path = os.path.join(self.model_dir, "tensorrt")
        check_make_directory(path)
        return path

    @staticmethod
    def get_tensorrt_clip_key(
        size: int, lora: List[Tuple[str, float]], inversion: List[str], **kwargs: Any
    ) -> str:
        """
        Uses hashlib to generate the unique key for the CLIP engine.
        CLIP must be rebuilt for each:
            1. Model
            2. Dimension
            3. LoRA
            4. Textual Inversion
        """
        return md5(
            "-".join(
                [
                    str(size),
                    ":".join(
                        "=".join([str(part) for part in lora_weight])
                        for lora_weight in sorted(lora, key=lambda lora_part: lora_part[0])
                    ),
                    ":".join(sorted(inversion)),
                ]
            ).encode("utf-8")
        ).hexdigest()

    @property
    def model_tensorrt_clip_key(self) -> str:
        """
        Gets the CLIP key for the current configuration.
        """
        return DiffusionPipelineManager.get_tensorrt_clip_key(
            size=self.size,
            lora=self.lora_names_weights,
            inversion=self.inversion_names,
        )

    @property
    def model_tensorrt_clip_dir(self) -> str:
        """
        Gets where the tensorrt CLIP engine will be stored.
        """
        path = os.path.join(self.model_tensorrt_dir, "clip", self.model_tensorrt_clip_key)
        check_make_directory(path)
        metadata_path = os.path.join(path, "metadata.json")
        if not os.path.exists(metadata_path):
            self.write_tensorrt_metadata(metadata_path)
        return path

    @staticmethod
    def get_tensorrt_unet_key(
        size: int,
        lora: List[Tuple[str, float]],
        inversion: List[str],
        **kwargs: Any,
    ) -> str:
        """
        Uses hashlib to generate the unique key for the UNET engine.
        UNET must be rebuilt for each:
            1. Model
            2. Dimension
            3. LoRA
            4. Textual Inversion
        """
        return md5(
            "-".join(
                [
                    str(size),
                    ":".join(
                        "=".join([str(part) for part in lora_weight])
                        for lora_weight in sorted(lora, key=lambda lora_part: lora_part[0])
                    ),
                    ":".join(sorted(inversion)),
                ]
            ).encode("utf-8")
        ).hexdigest()

    @property
    def model_tensorrt_unet_key(self) -> str:
        """
        Gets the UNET key for the current configuration.
        """
        return DiffusionPipelineManager.get_tensorrt_unet_key(
            size=self.size,
            lora=self.lora_names_weights,
            inversion=self.inversion_names,
        )

    @property
    def model_tensorrt_unet_dir(self) -> str:
        """
        Gets where the tensorrt CLIP engine will be stored.
        """
        path = os.path.join(self.model_tensorrt_dir, "unet", self.model_tensorrt_unet_key)
        check_make_directory(path)
        metadata_path = os.path.join(path, "metadata.json")
        if not os.path.exists(metadata_path):
            self.write_tensorrt_metadata(metadata_path)
        return path

    @staticmethod
    def get_tensorrt_controlled_unet_key(
        size: int,
        lora: List[Tuple[str, float]],
        inversion: List[str],
        **kwargs: Any,
    ) -> str:
        """
        Uses hashlib to generate the unique key for the UNET engine with controlnet blocks.
        ControlledUNET must be rebuilt for each:
            1. Model
            2. Dimension
            3. LoRA
            4. Textual Inversion
        """
        return md5(
            "-".join(
                [
                    str(size),
                    ":".join(
                        "=".join([str(part) for part in lora_weight])
                        for lora_weight in sorted(lora, key=lambda lora_part: lora_part[0])
                    ),
                    ":".join(sorted(inversion)),
                ]
            ).encode("utf-8")
        ).hexdigest()

    @property
    def model_tensorrt_controlled_unet_key(self) -> str:
        """
        Gets the UNET key for the current configuration.
        """
        return DiffusionPipelineManager.get_tensorrt_controlled_unet_key(
            size=self.size,
            lora=self.lora_names_weights,
            inversion=self.inversion_names,
        )

    @property
    def model_tensorrt_controlled_unet_dir(self) -> str:
        """
        Gets where the tensorrt Controlled UNet engine will be stored.
        """
        path = os.path.join(
            self.model_tensorrt_dir, "controlledunet", self.model_tensorrt_controlled_unet_key
        )
        check_make_directory(path)
        metadata_path = os.path.join(path, "metadata.json")
        if not os.path.exists(metadata_path):
            self.write_tensorrt_metadata(metadata_path)
        return path

    @staticmethod
    def get_tensorrt_controlnet_key(size: int, controlnet: str, **kwargs: Any) -> str:
        """
        Uses hashlib to generate the unique key for the controlnet engine.
        controlnet must be rebuilt for each:
            1. Model
            2. Dimension
            3. ControlNet
        """
        return md5("-".join([str(size), controlnet]).encode("utf-8")).hexdigest()

    @property
    def model_tensorrt_controlnet_key(self) -> str:
        """
        Gets the UNET key for the current configuration.
        """
        return DiffusionPipelineManager.get_tensorrt_controlnet_key(
            size=self.size,
            controlnet=""
            if self.controlnet is None
            else getattr(self.controlnet.config, "_name_or_path", ""),
        )

    @property
    def model_tensorrt_controlnet_dir(self) -> str:
        """
        Gets where the tensorrt controlnet engine will be stored.
        """
        path = os.path.join(
            self.model_tensorrt_dir, "controlnet", self.model_tensorrt_controlnet_key
        )
        check_make_directory(path)
        metadata_path = os.path.join(path, "metadata.json")
        if not os.path.exists(metadata_path):
            self.write_tensorrt_metadata(metadata_path)
        return path

    @staticmethod
    def get_tensorrt_vae_key(size: int, **kwargs: Any) -> str:
        """
        Uses hashlib to generate the unique key for the VAE engine.
        VAE must be rebuilt for each:
            1. Model
            2. Dimension
        """
        return md5(str(size).encode("utf-8")).hexdigest()

    @property
    def model_tensorrt_vae_key(self) -> str:
        """
        Gets the UNET key for the current configuration.
        """
        return DiffusionPipelineManager.get_tensorrt_vae_key(size=self.size)

    @property
    def model_tensorrt_vae_dir(self) -> str:
        """
        Gets where the tensorrt VAE engine will be stored.
        """
        path = os.path.join(self.model_tensorrt_dir, "vae", self.model_tensorrt_vae_key)
        check_make_directory(path)
        metadata_path = os.path.join(path, "metadata.json")
        if not os.path.exists(metadata_path):
            self.write_tensorrt_metadata(metadata_path)
        return path

    @property
    def tensorrt_is_supported(self) -> bool:
        """
        Tries to import tensorrt to see if it's supported.
        """
        if not hasattr(self, "_tensorrt_is_supported"):
            try:
                import tensorrt

                tensorrt.__version__  # quiet importchecker
                self._tensorrt_is_supported = True
            except Exception as ex:
                logger.info("TensorRT is disabled.")
                logger.debug("{0}: {1}".format(type(ex).__name__, ex))
                self._tensorrt_is_supported = False
        return self._tensorrt_is_supported

    @property
    def tensorrt_is_enabled(self) -> bool:
        """
        By default this is always enabled. This is independent from supported/ready.
        """
        if not hasattr(self, "_tensorrt_enabled"):
            self._tensorrt_enabled = True
        return self._tensorrt_enabled

    @tensorrt_is_enabled.setter
    def tensorrt_is_enabled(self, new_enabled: bool) -> None:
        """
        Disables or enables TensorRT.
        """
        if new_enabled != self.tensorrt_is_enabled:
            del self.pipeline
        self._tensorrt_enabled = new_enabled

    @property
    def tensorrt_is_ready(self) -> bool:
        """
        Checks to determine if Tensor RT is ready based on the existence of engines.
        """
        if not self.tensorrt_is_supported:
            return False
        from enfugue.diffusion.rt.engine import Engine

        trt_ready = True
        if "vae" in self.TENSORRT_STAGES:
            trt_ready = trt_ready and os.path.exists(
                Engine.get_engine_path(self.model_tensorrt_vae_dir)
            )
        if "clip" in self.TENSORRT_STAGES:
            trt_ready = trt_ready and os.path.exists(
                Engine.get_engine_path(self.model_tensorrt_clip_dir)
            )
        if self.controlnet is not None or self.TENSORRT_ALWAYS_USE_CONTROLLED_UNET:
            if "unet" in self.TENSORRT_STAGES:
                trt_ready = trt_ready and os.path.exists(
                    Engine.get_engine_path(self.model_tensorrt_controlled_unet_dir)
                )
            if "controlnet" in self.TENSORRT_STAGES:
                trt_ready = trt_ready and os.path.exists(
                    Engine.get_engine_path(self.model_tensorrt_controlnet_dir)
                )
        elif "unet" in self.TENSORRT_STAGES:
            trt_ready = trt_ready and os.path.exists(
                Engine.get_engine_path(self.model_tensorrt_unet_dir)
            )
        return trt_ready

    @property
    def build_tensorrt(self) -> bool:
        """
        Checks to see if TensorRT should be built based on configuration.
        """
        if not hasattr(self, "_build_tensorrt"):
            if not self.tensorrt_is_supported:
                self._build_tensorrt = False
            else:
                self._build_tensorrt = self.configuration.get("enfugue.tensorrt", False)
        return self._build_tensorrt

    @build_tensorrt.setter
    def build_tensorrt(self, new_build: bool) -> None:
        """
        Changes whether or not TensorRT engines should be built when absent
        """
        self._build_tensorrt = new_build
        if not self.tensorrt_is_ready and self.tensorrt_is_supported:
            del self.pipeline  # Prepare for build

    @property
    def use_tensorrt(self) -> bool:
        """
        Gets the ultimate decision on whether the tensorrt pipeline should be used.
        """
        return (self.tensorrt_is_ready or self.build_tensorrt) and self.tensorrt_is_enabled

    @property
    def pipeline_class(self) -> Type:
        """
        Gets the pipeline class to use.
        """
        if self.use_tensorrt:
            from enfugue.diffusion.rt.pipeline import EnfugueTensorRTStableDiffusionPipeline

            return EnfugueTensorRTStableDiffusionPipeline
        else:
            from enfugue.diffusion.pipeline import EnfugueStableDiffusionPipeline

            return EnfugueStableDiffusionPipeline

    @property
    def model(self) -> str:
        """
        Gets the configured model.
        """
        if not hasattr(self, "_model"):
            self._model = self.configuration.get("enfugue.model", DEFAULT_MODEL)
            if self._model.endswith(".ckpt") or self._model.endswith(".safetensors"):
                self._model = self.check_convert_checkpoint(self._model)
        return self._model

    @model.setter
    def model(self, new_model: Optional[str]) -> None:
        """
        Sets a new model. Destroys the pipeline.
        """
        if new_model is None:
            new_model = self.configuration.get("enfugue.model", DEFAULT_MODEL)
        if new_model.endswith(".ckpt") or new_model.endswith(".safetensors"):
            new_model = self.check_convert_checkpoint(new_model)
        if self.model != new_model:
            del self.pipeline
        self._model = new_model

    @property
    def dtype(self) -> torch.dtype:
        if not hasattr(self, "_torch_dtype"):
            if self.device == "cpu":
                logger.debug("Inferencing on CPU, using BFloat")
                self._torch_dtype = torch.float
            else:
                configuration_dtype = self.configuration.get("enfugue.dtype", "float16")
                if configuration_dtype == "float16" or configuration_dtype == "half":
                    self._torch_dtype = torch.half
                elif configuration_dtype == "float32" or configuration_dtype == "float":
                    self._torch_dtype = torch.float
                else:
                    raise ConfigurationError(
                        "dtype incorrectly configured, use 'float16/half' or 'float32/float'"
                    )
        return self._torch_dtype

    @dtype.setter
    def dtype(self, new_dtype: Union[str, torch.dtype]) -> None:
        """
        Sets the torch dtype.
        Deletes the pipeline if it's different from the previous one.
        """
        if self.device == "cpu":
            raise ValueError("CPU-based diffusion can only use bfloat")
        if new_dtype == "float16" or new_dtype == "half":
            new_dtype = torch.half
        elif new_dtype == "float32" or new_dtype == "float":
            new_dtype = torch.float
        else:
            raise ConfigurationError(
                "dtype incorrectly configured, use 'float16/half' or 'float32/float'"
            )

        if getattr(self, "_torch_dtype", new_dtype) != new_dtype:
            del self.pipeline

        self._torch_dtype = new_dtype

    @property
    def lora(self) -> List[Tuple[str, float]]:
        """
        Get LoRA added to the text encoder and UNet.
        """
        return getattr(self, "_lora", [])

    @lora.setter
    def lora(
        self, new_lora: Optional[Union[str, List[str], Tuple[str, float], List[Tuple[str, float]]]]
    ) -> None:
        """
        Sets new LoRA. Destroys the pipeline.
        """
        if new_lora is None:
            if hasattr(self, "_lora") and len(self._lora) > 0:
                del self.pipeline
            self._lora: List[Tuple[str, float]] = []
            return

        lora: List[Tuple[str, float]] = []
        if isinstance(new_lora, list):
            for this_lora in new_lora:
                if isinstance(this_lora, list):
                    lora.append(tuple(this_lora))  # type: ignore[unreachable]
                elif isinstance(this_lora, tuple):
                    lora.append(this_lora)
                else:
                    lora.append((this_lora, 1))
        elif isinstance(new_lora, tuple):
            lora = [new_lora]
        else:
            lora = [(new_lora, 1)]

        if getattr(self, "_lora", []) != lora:
            del self.pipeline
            self._lora = lora

    @property
    def lora_names_weights(self) -> List[Tuple[str, float]]:
        """
        Gets the basenames of any LoRA present.
        """
        return [(os.path.splitext(os.path.basename(lora))[0], weight) for lora, weight in self.lora]

    @property
    def inversion(self) -> List[str]:
        """
        Get textual inversions added to the text encoder.
        """
        return getattr(self, "_inversion", [])

    @inversion.setter
    def inversion(self, new_inversion: Optional[Union[str, List[str]]]) -> None:
        """
        Sets new textual inversions. Destroys the pipeline.
        """
        if new_inversion is None:
            if hasattr(self, "_inversion") and len(self._inversion) > 0:
                del self.pipeline
            self._inversion: List[str] = []
            return

        if not isinstance(new_inversion, list):
            new_inversion = [new_inversion]
        if getattr(self, "_inversion", []) != new_inversion:
            del self.pipeline
            self._inversion = new_inversion

    @property
    def inversion_names(self) -> List[str]:
        """
        Gets the basenames of any textual inversions present.
        """
        return [os.path.splitext(os.path.basename(inversion))[0] for inversion in self.inversion]

    @property
    def model_ckpt_path(self) -> str:
        """
        Gets the path for a model checkpoint. May be a .ckpt or .safetensors.
        """
        ckpt_path = os.path.join(self.engine_checkpoints_dir, f"{self.model}.ckpt")
        safetensor_path = os.path.join(self.engine_checkpoints_dir, f"{self.model}.safetensors")

        if os.path.exists(ckpt_path):
            return ckpt_path
        elif os.path.exists(safetensor_path):
            return safetensor_path
        else:
            raise IOError(f"Unknown model {self.model}")

    @property
    def vae_config(self) -> Dict:
        """
        Reads the VAE config.json file.
        """
        path = os.path.join(self.model_dir, "vae", "config.json")
        if not os.path.exists(path):
            raise OSError(f"Couldn't find VAE config file at {path}")
        return cast(Dict, load_json(path))

    @property
    def unet_config(self) -> Dict:
        """
        Reads the unet config.json file.
        """
        path = os.path.join(self.model_dir, "unet", "config.json")
        if not os.path.exists(path):
            raise OSError(f"Couldn't find unet config file at {path}")
        return cast(Dict, load_json(path))

    @property
    def clip_config(self) -> Dict:
        """
        Reads the clip config.json file.
        """
        path = os.path.join(self.model_dir, "text_encoder", "config.json")
        if not os.path.exists(path):
            raise OSError(f"Couldn't find clip config file at {path}")
        return cast(Dict, load_json(path))

    @property
    def inpainting(self) -> bool:
        """
        Returns true if the model is an inpainting model.
        """
        return self.unet_config["in_channels"] == 9

    @inpainting.setter
    def inpainting(self, new_inpainting: bool) -> None:
        """
        Sets whether or not we are inpainting.

        We trade efficiency for ease-of-use here; we just keep a model named `-inpainting`
        for any model.
        """
        if self.inpainting != new_inpainting:
            del self.pipeline
            current_checkpoint_path = self.model_ckpt_path

            default_checkpoint_name, _ = os.path.splitext(os.path.basename(DEFAULT_MODEL))
            default_inpainting_name, _ = os.path.splitext(
                os.path.basename(DEFAULT_INPAINTING_MODEL)
            )
            checkpoint_name, ext = os.path.splitext(os.path.basename(current_checkpoint_path))

            if default_checkpoint_name == checkpoint_name and new_inpainting:
                logger.info(f"Switching to default inpainting checkpoint")
                self.model = self.check_download_checkpoint(DEFAULT_INPAINTING_MODEL)
            elif default_inpainting_name == checkpoint_name and not new_inpainting:
                logger.info(f"Switching to default model")
                self.model = self.check_download_checkpoint(DEFAULT_MODEL)
            else:
                target_checkpoint_name = checkpoint_name

                if new_inpainting:
                    target_checkpoint_name += "-inpainting"
                else:
                    target_checkpoint_name = target_checkpoint_name[:-11]
                target_checkpoint_path = os.path.join(
                    os.path.dirname(current_checkpoint_path), f"{target_checkpoint_name}{ext}"
                )
                if not os.path.exists(target_checkpoint_path):
                    if not new_inpainting:
                        raise NotImplementedError(
                            "Creating a non-inpainting checkpoint from an inpainting one is not yet supported."
                        )
                    logger.info(f"Creating inpainting checkpoint from {current_checkpoint_path}")
                    self.create_inpainting_checkpoint(
                        current_checkpoint_path, target_checkpoint_path
                    )
                logger.info(
                    "Switching to {0}inpainting checkpoint at {1}".format(
                        "" if new_inpainting else "non-", target_checkpoint_path
                    )
                )
                self.model = target_checkpoint_path

    @property
    def pipeline(self) -> EnfugueStableDiffusionPipeline:
        """
        Instantiates the pipeline.
        """
        if not hasattr(self, "_pipeline"):
            kwargs = {
                "cache_dir": self.diffusers_cache_dir,
                "engine_size": self.size,
                "chunking_size": self.chunking_size,
                "requires_safety_checker": self.safe,
                "torch_dtype": self.dtype,
            }

            if not self.safe:
                kwargs["safety_checker"] = None

            if self.use_tensorrt:
                if "unet" in self.TENSORRT_STAGES:
                    if self.controlnet is None and not self.TENSORRT_ALWAYS_USE_CONTROLLED_UNET:
                        kwargs["unet_engine_dir"] = self.model_tensorrt_unet_dir
                    else:
                        kwargs[
                            "controlled_unet_engine_dir"
                        ] = self.model_tensorrt_controlled_unet_dir
                if "controlnet" in self.TENSORRT_STAGES and self.controlnet is not None:
                    kwargs["controlnet_engine_dir"] = self.model_tensorrt_controlnet_dir
                if "vae" in self.TENSORRT_STAGES:
                    kwargs["vae_engine_dir"] = self.model_tensorrt_vae_dir
                if "clip" in self.TENSORRT_STAGES:
                    kwargs["clip_engine_dir"] = self.model_tensorrt_clip_dir

            logger.debug(
                f"Initializing pipeline model {self.model} in directory {self.model_dir} with arguments {kwargs}"
            )
            kwargs["controlnet"] = self.controlnet

            pipeline = self.pipeline_class.from_pretrained(self.model_dir, **kwargs)
            if not self.tensorrt_is_ready:
                for lora, weight in self.lora:
                    logger.debug(f"Adding LoRA {lora} to pipeline")
                    pipeline.load_lora_weights(lora, multiplier=weight)
                for inversion in self.inversion:
                    logger.debug(f"Adding textual inversion {inversion} to pipeline")
                    pipeline.load_textual_inversion(inversion)
            self._pipeline = pipeline.to(self.device)
        return self._pipeline

    @pipeline.deleter
    def pipeline(self) -> None:
        """
        Eliminates any instantiated pipeline.
        """
        if hasattr(self, "_pipeline"):
            logger.debug("Deleting pipeline.")
            del self._pipeline
            import torch

            torch.cuda.empty_cache()
        else:
            logger.debug("Pipeline delete called, but no pipeline present.")

    def unload_pipeline(self) -> None:
        """
        Calls the pipeline deleter.
        """
        del self.pipeline

    @property
    def upscaler(self) -> Upscaler:
        """
        Gets the GAN upscaler
        """
        if not hasattr(self, "_upscaler"):
            from enfugue.diffusion.upscale import Upscaler

            self._upscaler = Upscaler(self.engine_other_dir)
        return self._upscaler

    @upscaler.deleter
    def upscaler(self) -> None:
        """
        Deletes the upscaler to save VRAM.
        """
        if hasattr(self, "_upscaler"):
            logger.debug("Deleting upscaler.")
            del self._upscaler
            import torch

            torch.cuda.empty_cache()

    @property
    def edge_detector(self) -> EdgeDetector:
        """
        Gets the edge detector.
        """
        if not hasattr(self, "_edge_detector"):
            from enfugue.diffusion.edge.detect import EdgeDetector

            self._edge_detector = EdgeDetector(self.engine_other_dir)
        return self._edge_detector

    def unload_upscaler(self) -> None:
        """
        Calls the upscaler deleter.
        """
        del self.upscaler

    def get_controlnet(self, controlnet: Optional[str] = None) -> Optional[ControlNetModel]:
        """
        Loads a controlnet
        """
        if controlnet is None:
            return None
        from diffusers.models import ControlNetModel

        self.start_keepalive()
        result = ControlNetModel.from_pretrained(
            controlnet,
            torch_dtype=torch.half,
            cache_dir=self.diffusers_cache_dir,
        )
        self.stop_keepalive()
        return result

    @property
    def controlnet(self) -> Optional[ControlNetModel]:
        """
        Gets the configured controlnet (or none.)
        """
        if not hasattr(self, "_controlnet"):
            self._controlnet = self.get_controlnet(self.controlnet_name)
        return self._controlnet

    @controlnet.setter
    def controlnet(
        self,
        new_controlnet: Optional[Literal["canny", "tile", "mlsd", "hed", "scribble", "inpaint"]],
    ) -> None:
        """
        Sets a new controlnet.
        """
        pretrained_path = None
        if new_controlnet == "canny":
            pretrained_path = CONTROLNET_CANNY
        elif new_controlnet == "mlsd":
            pretrained_path = CONTROLNET_MLSD
        elif new_controlnet == "hed":
            pretrained_path = CONTROLNET_HED
        elif new_controlnet == "tile":
            pretrained_path = CONTROLNET_TILE
        elif new_controlnet == "scribble":
            pretrained_path = CONTROLNET_SCRIBBLE
        elif new_controlnet == "inpaint":
            pretrained_path = CONTROLNET_INPAINT
        if pretrained_path is None and new_controlnet is not None:
            logger.error(f"Unsupported controlnet {new_controlnet}")

        existing_controlnet = getattr(self, "_controlnet", None)

        if (
            (existing_controlnet is None and new_controlnet is not None)
            or (existing_controlnet is not None and new_controlnet is None)
            or (existing_controlnet is not None and self.controlnet_name != new_controlnet)
        ):
            del self.pipeline
            if new_controlnet is not None:
                self._controlnet_name = new_controlnet
                self._controlnet = self.get_controlnet(pretrained_path)
            else:
                self._controlnet_name = None  # type: ignore
                self._controlnet = None

    @property
    def controlnet_name(self) -> Optional[str]:
        """
        Gets the name of the control net, if one was set.
        """
        if not hasattr(self, "_controlnet_name"):
            self._controlnet_name = self.configuration.get("enfugue.controlnet", None)
        return self._controlnet_name

    def check_download_checkpoint(self, remote_url: str) -> str:
        """
        Downloads a checkpoint directly to the checkpoints folder.
        """
        output_file = os.path.basename(remote_url)
        output_path = os.path.join(self.engine_checkpoints_dir, output_file)
        self.start_keepalive()
        check_download(remote_url, output_path)
        self.stop_keepalive()
        return output_path

    def check_convert_checkpoint(self, checkpoint_path: str) -> str:
        """
        Converts a .ckpt file to the directory structure from diffusers
        """
        checkpoint_file = os.path.basename(checkpoint_path)
        model_name, ext = os.path.splitext(checkpoint_file)
        model_dir = os.path.join(self.engine_models_dir, model_name)

        if not os.path.exists(model_dir):
            if checkpoint_path.startswith("http"):
                checkpoint_path = self.check_download_checkpoint(checkpoint_path)

            from diffusers.pipelines.stable_diffusion.convert_from_ckpt import (
                download_from_original_stable_diffusion_ckpt,
            )

            self.start_keepalive()
            pipe = download_from_original_stable_diffusion_ckpt(
                checkpoint_path=checkpoint_path,
                scheduler_type="ddim",
                from_safetensors=ext == ".safetensors",
                num_in_channels=9 if "inpaint" in checkpoint_path else 4,
            ).to(torch_dtype=self.dtype)
            pipe.save_pretrained(model_dir)
            self.stop_keepalive()
        return model_name

    def __call__(self, **kwargs: Any) -> Any:
        """
        Passes an invocation down to the pipeline, doing whatever it needs to do to initialize it.
        Will switch between inpainting and non-inpainting models
        """
        logger.debug(f"Calling pipeline with arguments {kwargs}")
        if kwargs.get("mask", None) is not None:
            self.inpainting = True
        else:
            self.inpainting = False
        called_width = kwargs.get("width", self.size)
        called_height = kwargs.get("height", self.size)
        chunk_size = kwargs.get("chunking_size", self.chunking_size)
        if called_width < self.size:
            self.tensorrt_is_enabled = False
            logger.info(
                f"Width ({called_width}) less than configured width ({self.size}), disabling TensorRT"
            )
        elif called_height < self.size:
            logger.info(
                f"Height ({called_height}) less than configured height ({self.size}), disabling TensorRT"
            )
            self.tensorrt_is_enabled = False
        elif (called_width != self.size or called_height != self.size) and not chunk_size:
            logger.info(
                f"Dimensions do not match size of engine and chunking is disabled, disabling TensorRT"
            )
            self.tensorrt_is_enabled = False
        else:
            self.tenssort_is_enabled = True
        result = self.pipeline(generator=self.generator, **kwargs)
        self.tensorrt_is_enabled = True
        return result

    def write_tensorrt_metadata(self, path: str) -> None:
        """
        Writes metadata for TensorRT to a json file
        """
        if "controlnet" in path:
            dump_json(path, {"size": self.size, "controlnet": self.controlnet_name})
        else:
            dump_json(
                path,
                {
                    "size": self.size,
                    "lora": self.lora_names_weights,
                    "inversion": self.inversion_names,
                },
            )

    @staticmethod
    def get_tensorrt_status(
        engine_root: str,
        model: str,
        size: Optional[int] = None,
        lora: Optional[Union[str, Tuple[str, float], List[Union[str, Tuple[str, float]]]]] = None,
        inversion: Optional[Union[str, List[str]]] = None,
        controlnet: Optional[Union[str, List[str]]] = None,
    ) -> Dict[str, bool]:
        """
        Gets the TensorRT status for an individual model.
        """
        tensorrt_is_supported = False

        try:
            import tensorrt

            tensorrt.__version__  # quiet importchecker
            tensorrt_is_supported = True
        except Exception as ex:
            logger.info("TensorRT is disabled.")
            logger.debug("{0}: {1}".format(type(ex).__name__, ex))
            pass

        if not tensorrt_is_supported:
            return {"supported": False, "ready": False}

        if model.endswith(".ckpt") or model.endswith(".safetensors"):
            model, _ = os.path.splitext(os.path.basename(model))
        else:
            model = os.path.basename(model)

        if model.endswith("-inpainting"):
            # Look for the base model instead, we'll look for inpainting separately
            model, _, _ = model.rpartition("-")

        model_dir = os.path.join(engine_root, "models", model)
        inpaint_model_dir = os.path.join(engine_root, "models", f"{model}-inpainting")

        if size is None:
            size = DiffusionPipelineManager.DEFAULT_SIZE

        if inversion is None:
            inversion = []
        elif not isinstance(inversion, list):
            inversion = [inversion]

        inversion_key = []
        for inversion_part in inversion:
            inversion_name, ext = os.path.splitext(os.path.basename(inversion_part))
            inversion_key.append(inversion_name)

        if lora is None:
            lora = []
        elif not isinstance(lora, list):
            lora = [lora]

        lora_key = []
        for lora_part in lora:
            if isinstance(lora_part, tuple):
                lora_path, lora_weight = lora_part
            else:
                lora_path, lora_weight = lora_part, 1.0
            lora_name, ext = os.path.splitext(os.path.basename(lora_path))
            lora_key.append((lora_name, lora_weight))

        clip_ready = "clip" not in DiffusionPipelineManager.TENSORRT_STAGES
        vae_ready = "vae" not in DiffusionPipelineManager.TENSORRT_STAGES
        unet_ready = "unet" not in DiffusionPipelineManager.TENSORRT_STAGES
        inpaint_unet_ready = unet_ready
        controlled_unet_ready = unet_ready
        controlnet_ready: Union[bool, Dict[str, bool]] = (
            "controlnet" not in DiffusionPipelineManager.TENSORRT_STAGES
        )

        if not clip_ready:
            clip_key = DiffusionPipelineManager.get_tensorrt_clip_key(
                size, lora=lora_key, inversion=inversion_key
            )
            clip_plan = os.path.join(model_dir, "tensorrt", "clip", clip_key, "engine.plan")
            clip_ready = os.path.exists(clip_plan)

        if not vae_ready:
            vae_key = DiffusionPipelineManager.get_tensorrt_vae_key(
                size, lora=lora_key, inversion=inversion_key
            )
            vae_plan = os.path.join(model_dir, "tensorrt", "vae", vae_key, "engine.plan")
            vae_ready = os.path.exists(vae_plan)

        if not unet_ready:
            unet_key = DiffusionPipelineManager.get_tensorrt_unet_key(
                size, lora=lora_key, inversion=inversion_key
            )
            unet_plan = os.path.join(model_dir, "tensorrt", "unet", unet_key, "engine.plan")
            unet_ready = os.path.exists(unet_plan)

            inpaint_unet_plan = os.path.join(
                inpaint_model_dir, "tensorrt", "unet", unet_key, "engine.plan"
            )
            inpaint_unet_ready = os.path.exists(inpaint_unet_plan)

            controlled_unet_key = DiffusionPipelineManager.get_tensorrt_controlled_unet_key(
                size, lora=lora_key, inversion=inversion_key
            )
            controlled_unet_plan = os.path.join(
                model_dir, "tensorrt", "controlledunet", controlled_unet_key, "engine.plan"
            )
            controlled_unet_ready = os.path.exists(controlled_unet_plan)

        if not controlnet_ready:
            if controlnet is None:
                controlnet_ready = True
            else:
                controlnet_ready = {}
                if not isinstance(controlnet, list):
                    controlnet = [controlnet]
                for controlnet_name in controlnet:
                    controlnet_key = DiffusionPipelineManager.get_tensorrt_controlnet_key(
                        size, controlnet=controlnet_name
                    )
                    controlnet_plan = os.path.join(
                        model_dir, "tensorrt", "controlnet", controlnet_key, "engine.plan"
                    )
                    controlnet_ready[controlnet_name] = os.path.exists(controlnet_plan)

        ready = clip_ready and vae_ready
        if controlnet is not None or DiffusionPipelineManager.TENSORRT_ALWAYS_USE_CONTROLLED_UNET:
            ready = ready and controlled_unet_ready
            if isinstance(controlnet_ready, dict):
                for name in controlnet_ready:
                    ready = ready and controlnet_ready[name]
        else:
            ready = ready and unet_ready

        return {
            "supported": tensorrt_is_supported,
            "unet_ready": unet_ready,
            "controlled_unet_ready": controlled_unet_ready,
            "vae_ready": vae_ready,
            "clip_ready": clip_ready,
            "inpaint_unet_ready": inpaint_unet_ready,
            "ready": ready,
        }

    def create_inpainting_checkpoint(
        self, source_checkpoint_path: str, target_checkpoint_path: str
    ) -> None:
        """
        Creates an inpainting model by merging in the SD 1.5 inpainting model with a non inpainting model.
        """
        from enfugue.diffusion.util import ModelMerger

        self.start_keepalive()
        merger = ModelMerger(
            self.check_download_checkpoint(DEFAULT_INPAINTING_MODEL),
            source_checkpoint_path,
            self.check_download_checkpoint(DEFAULT_MODEL),
            interpolation="add-difference",
        )
        merger.save(target_checkpoint_path)
        logger.info(f"Saved merged inpainting checkpoint at {target_checkpoint_path}")
        self.stop_keepalive()



"""
This is the entry point for the windows .exe
"""
if __name__ == "__main__":
    import multiprocessing

    multiprocessing.freeze_support()  # pyinstaller fix

    from enfugue.util import get_local_configuration
    import os
    import platform

    os.environ["CUDA_MODULE_LOADING"] = "LAZY"

    from enfugue.server import EnfugueServer

    system = platform.system()
    configuration = get_local_configuration()
    if system == "Windows":
        EnfugueServer.serve_icon(configuration)
    else:
        server = EnfugueServer()
        server.configure(**configuration)
        server.serve()



from enfugue.util.downloads import *
from enfugue.util.gpu import *
from enfugue.util.images import *
from enfugue.util.installation import *
from enfugue.util.log import *
from enfugue.util.log import *
from enfugue.util.signature import *
from enfugue.util.tokens import *



from typing import Optional, List

from polygraphy.backend.onnx.loader import fold_constants

from onnx_graphsurgeon import import_onnx, export_onnx

from onnx_graphsurgeon.ir.graph import Graph
from onnx.onnx_ml_pb2 import ModelProto as ONNXModel
from onnx.shape_inference import infer_shapes


class Optimizer:
    graph: Graph

    def __init__(self, onnx_model: ONNXModel) -> None:
        self.graph = import_onnx(onnx_model)

    def select_outputs(self, to_select: List[int], names: Optional[List[str]] = None) -> None:
        """
        Selects and reindexes  graph outputs.
        """
        self.graph.outputs = [self.graph.outputs[i] for i in to_select]
        if names:
            for i, name in enumerate(names):
                self.graph.outputs[i].name = name

    def cleanup(self, return_onnx: bool = False) -> Optional[ONNXModel]:
        """
        Cleans up the graph and re-sorts.
        """
        self.graph.cleanup().toposort()
        if return_onnx:
            return export_onnx(self.graph)
        return None

    def fold_constants(self, return_onnx: bool = False) -> Optional[ONNXModel]:
        """
        Folds the constants from the graph, then re-imports.
        """
        onnx_graph = fold_constants(export_onnx(self.graph), allow_onnxruntime_shape_inference=True)
        self.graph = import_onnx(onnx_graph)
        if return_onnx:
            return onnx_graph
        return None

    def infer_shapes(self, return_onnx: bool = False) -> Optional[ONNXModel]:
        """
        Runs shape inference on the graph and re-imports.
        """
        onnx_graph = export_onnx(self.graph)
        if onnx_graph.ByteSize() > 2147483648:
            raise TypeError("ERROR: model size exceeds supported 2GB limit")
        else:
            onnx_graph = infer_shapes(onnx_graph)

        self.graph = import_onnx(onnx_graph)
        if return_onnx:
            return onnx_graph
        return None

    @staticmethod
    def run(
        onnx_graph: ONNXModel,
        select_inputs: Optional[List[int]] = None,
        select_outputs: Optional[List[int]] = None,
        select_output_names: Optional[List[str]] = None,
    ) -> ONNXModel:
        """
        Creates an optimizer, runs it, and returns the optimized model
        """
        optimizer = Optimizer(onnx_graph)
        if select_inputs:
            optimizer.select_outputs(select_inputs)
        optimizer.cleanup()
        optimizer.fold_constants()
        optimizer.infer_shapes()
        if select_outputs:
            optimizer.select_outputs(select_outputs, select_output_names)
        return optimizer.cleanup(return_onnx=True)  # type: ignore



import os
import sys
import tqdm
import click
import termcolor
import logging
import traceback

from typing import Optional
from PIL import Image
from pibble.util.strings import Serializer
from pibble.util.log import LevelUnifiedLoggingContext
from pibble.api.configuration import APIConfiguration
from enfugue.util import logger


TRT_PIPELINE = "enfugue.diffusion.rt.pipeline.EnfugueTensorRTStableDiffusionPipeline"


@click.group(name="enfugue")
def main() -> None:
    """
    Enfugue command-line tools.
    """
    pass


@main.command(short_help="Prints the version of the main enfugue package and dependant packages.")
def version() -> None:
    """
    Gets the version of packages installed in the environment.
    """
    from enfugue.util import get_version

    try:
        enfugue_version = get_version()
    except:
        enfugue_version = "development"
    import torch

    click.echo(f"Enfugue v.{enfugue_version}")
    click.echo(f"Torch v.{torch.__version__}")
    click.echo("CUDA {0}".format("available" if torch.cuda.is_available() else "unavailable"))
    try:
        import tensorrt

        click.echo("TensorRT Supported")
    except ImportError:
        click.echo("TensorRT Unsupported")


@main.command(short_help="Runs the server.")
def run() -> None:
    """
    Runs the server synchronously using cherrypy.
    """
    from enfugue.util import get_local_configuration
    from enfugue.server import EnfugueServer

    configuration = get_local_configuration()
    server = EnfugueServer()
    server.configure(**configuration)
    server.serve()


try:
    main()
except Exception as ex:
    print(termcolor.colored(str(ex), "red"))
    if "--verbose" in sys.argv or "-v" in sys.argv:
        print(termcolor.colored(traceback.format_exc(), "red"))
    sys.exit(5)
sys.exit(0)



from __future__ import annotations

import os
import re
import numpy as np
import torch
import safetensors.torch

from typing import Dict, Type, Union, Optional, Literal, cast
from enfugue.util import logger

__all__ = ["DTypeConverter", "ModelMerger"]


class DTypeConverter:
    """
    This class converts between numpy and torch types.
    """

    numpy_to_torch: Dict[Type, torch.dtype] = {
        np.uint8: torch.uint8,
        np.int8: torch.int8,
        np.int16: torch.int16,
        np.int32: torch.int32,
        np.int64: torch.int64,
        np.float16: torch.float16,
        np.float32: torch.float32,
        np.float64: torch.float64,
        np.complex64: torch.complex64,
        np.complex128: torch.complex128,
        np.bool_: torch.bool,
    }

    torch_to_numpy: Dict[torch.dtype, Type] = {
        torch.uint8: np.uint8,
        torch.int8: np.int8,
        torch.int16: np.int16,
        torch.int32: np.int32,
        torch.int64: np.int64,
        torch.float16: np.float16,
        torch.float32: np.float32,
        torch.float64: np.float64,
        torch.complex64: np.complex64,
        torch.complex128: np.complex128,
        torch.bool: np.bool_,
    }

    @staticmethod
    def from_torch(torch_type: torch.dtype) -> Type:
        """
        Gets the numpy type from torch.
        :raises: KeyError When type is unknown.
        """
        return DTypeConverter.torch_to_numpy[torch_type]

    @staticmethod
    def from_numpy(numpy_type: Type) -> torch.dtype:
        """
        Gets the torch type from nump.
        :raises: KeyError When type is unknown.
        """
        return DTypeConverter.numpy_to_torch[numpy_type]

    @staticmethod
    def __call__(type_to_convert: Union[torch.dtype, Type]) -> Union[torch.dtype, Type]:
        """
        Converts from one type to the other, inferring based on the type passed.
        """
        if isinstance(type_to_convert, torch.dtype):
            return DTypeConverter.from_torch(type_to_convert)
        return DTypeConverter.from_numpy(type_to_convert)


class ModelMerger:
    """
    Allows merging various Stable Diffusion models of various sizes.
    Inspired by https://github.com/AUTOMATIC1111/stable-diffusion-webui/blob/master/modules/extras.py
    """

    CHECKPOINT_DICT_REPLACEMENTS = {
        "cond_stage_model.transformer.embeddings.": "cond_stage_model.transformer.text_model.embeddings.",
        "cond_stage_model.transformer.encoder.": "cond_stage_model.transformer.text_model.encoder.",
        "cond_stage_model.transformer.final_layer_norm.": "cond_stage_model.transformer.text_model.final_layer_norm.",
    }

    CHECKPOINT_DICT_SKIP = ["cond_stage_model.transformer.text_model.embeddings.position_ids"]

    discard_weights: Optional[re.Pattern]

    def __init__(
        self,
        primary_model: str,
        secondary_model: Optional[str],
        tertiary_model: Optional[str],
        interpolation: Optional[Literal["add-difference", "weighted-sum"]] = None,
        multiplier: Union[int, float] = 1.0,
        half: bool = True,
        discard_weights: Optional[Union[str, re.Pattern]] = None,
    ):
        self.primary_model = primary_model
        self.secondary_model = secondary_model
        self.tertiary_model = tertiary_model
        self.interpolation = interpolation
        self.multiplier = multiplier
        self.half = half
        if type(discard_weights) is str:
            self.discard_weights = re.compile(discard_weights)
        else:
            self.discard_weights = cast(Optional[re.Pattern], discard_weights)

    @staticmethod
    def as_half(tensor: torch.Tensor) -> torch.Tensor:
        """
        Halves a tensor if necessary
        """
        if tensor.dtype == torch.float:
            return tensor.half()
        return tensor

    @staticmethod
    def get_difference(theta0: torch.Tensor, theta1: torch.Tensor) -> torch.Tensor:
        """
        Simply gets the difference from two values.
        """
        return theta0 - theta1

    @staticmethod
    def weighted_sum(
        theta0: torch.Tensor, theta1: torch.Tensor, alpha: Union[int, float]
    ) -> torch.Tensor:
        """
        Returns the sum of θ0 and θ0 weighted by ɑ
        """
        return ((1 - alpha) * theta0) + (alpha * theta1)

    @staticmethod
    def add_weighted_difference(
        theta0: torch.Tensor, theta1: torch.Tensor, alpha: Union[int, float]
    ) -> torch.Tensor:
        """
        Adds a weighted difference back to the original value
        """
        return theta0 + (alpha * theta1)

    @staticmethod
    def get_state_dict_from_checkpoint(checkpoint: Dict) -> Dict:
        """
        Reads the raw state dictionary and find the proper state dictionary.
        """
        state_dict = checkpoint.pop("state_dict", checkpoint)
        if "state_dict" in state_dict:
            del state_dict["state_dict"]  # Remove any sub-embedded state dicts

        transformed_dict = dict(
            [
                (ModelMerger.transform_checkpoint_key(key), value)
                for key, value in state_dict.items()
            ]
        )
        state_dict.clear()
        state_dict.update(transformed_dict)
        return state_dict

    @staticmethod
    def load_checkpoint(checkpoint_path: str) -> Dict:
        """
        Loads a checkpoint"s state dictionary on the CPU for model merging.
        """
        _, ext = os.path.splitext(checkpoint_path)
        logger.debug(f"Model merger loading {checkpoint_path}")
        if ext.lower() == ".safetensors":
            ckpt = safetensors.torch.load_file(checkpoint_path, device="cpu")
        else:
            ckpt = torch.load(checkpoint_path, map_location="cpu")

        return ModelMerger.get_state_dict_from_checkpoint(ckpt)

    @staticmethod
    def is_ignored_key(key: str) -> bool:
        """
        Checks if a key should be ignored during merge.
        """
        return "model" not in key or key in ModelMerger.CHECKPOINT_DICT_SKIP

    @staticmethod
    def transform_checkpoint_key(text: str) -> str:
        """
        Transform a checkpoint key, if needed.
        """
        for key, value in ModelMerger.CHECKPOINT_DICT_REPLACEMENTS.items():
            if key.startswith(text):
                text = value + text[len(key) :]
        return text

    def save(self, output_path: str) -> None:
        """
        Runs the configured merger.
        """
        logger.debug(
            f"Executing model merger with interpolation '{self.interpolation}', primary model {self.primary_model}, secondary model {self.secondary_model}, tertiary model {self.tertiary_model}"
        )

        primary_state_dict = self.load_checkpoint(self.primary_model)
        secondary_state_dict = (
            None if not self.secondary_model else self.load_checkpoint(self.secondary_model)
        )
        tertiary_state_dict = (
            None if not self.tertiary_model else self.load_checkpoint(self.tertiary_model)
        )

        theta_0 = primary_state_dict
        theta_1 = secondary_state_dict

        if self.interpolation == "add-difference":
            if theta_1 is None or tertiary_state_dict is None:
                raise ValueError(f"{self.interpolation} requires three models.")
            logger.debug("Merging secondary and tertiary models.")
            for key in theta_1.keys():
                if self.is_ignored_key(key):
                    continue
                if key in tertiary_state_dict:
                    theta_1[key] = self.get_difference(theta_1[key], tertiary_state_dict[key])
                else:
                    theta_1[key] = torch.zeros_like(theta_1[key])
            del tertiary_state_dict

        if self.interpolation == "add-difference":
            interpolate = self.add_weighted_difference
        else:
            interpolate = self.weighted_sum

        if theta_1 is not None:
            logger.debug("Merging primary and secondary models.")
            for key in theta_0.keys():
                if key not in theta_1 or self.is_ignored_key(key):
                    continue

                a = theta_0[key]
                b = theta_1[key]

                # Check if we"re merging 4-channel (standard), 8-channel (ip2p) ir 9-channel (inpainting) model(s)
                if a.shape != b.shape and a.shape[0:1] + a.shape[2:] == b.shape[0:1] + b.shape[2:]:
                    if a.shape[1] == 4 and b.shape[1] == 9:
                        raise RuntimeError(
                            "When merging an inpainting model with a standard one, the primary model must be the inpainting model."
                        )
                    if a.shape[1] == 4 and b.shape[1] == 8:
                        raise RuntimeError(
                            "When merging an instruct-pix2pix model with a standard one, the primary model must be the instruct-pix2pix model."
                        )

                    if a.shape[1] == 8 and b.shape[1] == 4:
                        # Merging IP2P into Normal
                        theta_0[key][:, 0:4, :, :] = interpolate(
                            a[:, 0:4, :, :], b, self.multiplier
                        )
                        result_is_instruct_pix2pix_model = True
                    else:
                        # Merging inpainting into Normal
                        assert (
                            a.shape[1] == 9 and b.shape[1] == 4
                        ), f"Bad dimensions for merged layer {key}: A={a.shape}, B={b.shape}"
                        theta_0[key][:, 0:4, :, :] = interpolate(
                            a[:, 0:4, :, :], b, self.multiplier
                        )
                        result_is_inpainting_model = True
                else:
                    theta_0[key] = interpolate(a, b, self.multiplier)

                if self.half:
                    theta_0[key] = self.as_half(theta_0[key])

            del theta_1

        if self.discard_weights is not None:
            for key in list(theta_0):
                if re.search(self.discard_weights, key):
                    theta_0.pop(key, None)

        logger.debug(f"Saving merged model to {output_path}")
        _, extension = os.path.splitext(output_path)
        if extension.lower() == ".safetensors":
            safetensors.torch.save_file(theta_0, output_path)
        else:
            torch.save(theta_0, output_path)



from enfugue.database.base import EnfugueObjectBase
from pibble.database.orm import ORMVariadicType
from sqlalchemy import Column, String


class ConfigurationItem(EnfugueObjectBase):
    __tablename__ = "configuration_item"

    configuration_key = Column(String(256), primary_key=True)
    configuration_value = Column(ORMVariadicType(), nullable=True)



from enfugue.database.base import EnfugueObjectBase
from pibble.database.orm import ORMVariadicType
from pibble.ext.user.database import User
from sqlalchemy import Column, String, DateTime, Integer
from sqlalchemy.sql import func


class DiffusionInvocation(EnfugueObjectBase):
    __tablename__ = "invocation"

    id = Column(String(32), primary_key=True)
    user_id = Column(User.ForeignKey("id", ondelete="SET NULL", onupdate="SET NULL"), nullable=True)
    started = Column(DateTime(timezone=True), default=func.now(), nullable=False)
    plan = Column(ORMVariadicType(), nullable=False)
    duration = Column(Integer(), nullable=False, default=0)
    outputs = Column(Integer(), nullable=False, default=0)
    error = Column(String(512), nullable=True)

    user = User.Relationship(backref="invocations")



import os
from setuptools import setup, find_packages

package_name = "enfugue"
version_major = "0"
version_minor = "1"
version_patch = "1"

install_requires = [
    "cheroot>=9.0.0",
    "nvidia-pyindex>=1.0.9",
    "pibble[cherrypy]>=0.5",
    "torch>=1.13.1,<2.0",
    "numpy>=1.24.0",
    "colored>=1.4,<1.5",
    "diffusers>=0.16,<0.17",
    "albumentations>=0.4.3,<0.5",
    "opencv-python>=4.7.0.72,<4.8",
    "pudb==2019.2",
    "invisible-watermark>=0.1,<0.2",
    "imageio>=2.9,<3.0",
    "imageio-ffmpeg>=0.4.2,<0.5",
    "pytorch-lightning>=2.0.2,<2.1",
    "omegaconf>=2.1.1,<2.2",
    "test-tube>=0.7.5,<0.8",
    "streamlit>=0.73,<0.74",
    "einops>=0.3,<0.4",
    "torch-fidelity>=0.3,<0.4",
    "transformers>=4.28,<5.0",
    "torchmetrics==0.11.4",
    "kornia>=0.6,<0.7",
    "polygraphy>=0.47,<0.48",
    "accelerate>=0.18,<0.19",
    "tqdm>=4.27",
    "safetensors>=0.3,<0.4",
    "realesrgan>=0.3,<0.4",
    "gfpgan>=1.3.8,<1.4",
    "backgroundremover>=0.2.3,<0.3",
    "beautifulsoup4>=4.12,<5",
    "pystray>=0.19,<0.24",
    "html2text",
    "taming-transformers",
    "clip",
    "latent-diffusion",
]

extras_require = {
    "tensorrt": [
        "onnx==1.12",
        "onnxruntime-gpu==1.12.1",
        "onnx-graphsurgeon==0.3.26",
        "tensorrt>=8.6,<8.7",
    ],
    "build": [
        "types-protobuf>=4.23.0.1,<5.0",
        "types-requests>=2.30,<3.0",
        "types-setuptools>=67.7,<68.0",
        "types-urllib3>=1.26.25,<2.0",
        "types-tabulate>=0.9,<0.10",
        "importchecker>=2.0,<3.0",
    ],
}

packaged_files = []

here = os.path.dirname(os.path.abspath(__file__))
package_dir = os.path.join(here, package_name)
data_dirs = ["static", "config"]
for data_dir in data_dirs:
    data_dir_path = os.path.join(package_dir, data_dir)
    for dir_path, sub_dirs, file_names in os.walk(data_dir_path):
        for file_name in file_names:
            packaged_files.append(os.path.join(dir_path, file_name))

package_data = [os.path.relpath(packaged_file, package_dir) for packaged_file in packaged_files]

setup(
    name=package_name,
    author="Benjamin Paine",
    author_email="painebenjamin@gmail.com",
    version=f"{version_major}.{version_minor}.{version_patch}",
    packages=find_packages("."),
    package_data={package_name: ["py.typed", *package_data]},
    license="agpl-3.0",
    entry_points={"console_scripts": ["enfugue = enfugue.__main__:main"]},
    install_requires=install_requires,
    extras_require=extras_require,
)



import torch
from typing import Dict, List, Tuple
from enfugue.diffusion.rt.model.base import BaseModel


class ControlNet(BaseModel):
    """
    The controlled UNet uses a controlnet that forwards mid/down blocks.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        use_fp16: bool = True,
        device: str = "cuda",
        max_batch_size: int = 16,
        embedding_dim: int = 768,
        text_maxlen: int = 77,
        unet_dim: int = 4,
    ) -> None:
        super(ControlNet, self).__init__(
            model=model,
            use_fp16=use_fp16,
            device=device,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
            text_maxlen=text_maxlen,
        )
        self.unet_dim = unet_dim
        self.name = "ControlNet"

    def get_model_key(self) -> str:
        """
        Gets the unique name for this model.
        """
        return "controlnet"

    def get_input_names(self) -> List[str]:
        """
        Gets all input names of all kinds.
        """
        return [
            "sample",
            "timestep",
            "encoder_hidden_states",
            "controlnet_cond",
            "conditioning_scale",
        ]

    def get_output_names(self) -> List[str]:
        """
        Gets all output names of all kinds.
        """
        return [
            "mid_block",
            "down_block_0",
            "down_block_1",
            "down_block_2",
            "down_block_3",
            "down_block_4",
            "down_block_5",
            "down_block_6",
            "down_block_7",
            "down_block_8",
            "down_block_9",
            "down_block_10",
            "down_block_11",
        ]

    def get_dynamic_axes(self) -> Dict[str, Dict[int, str]]:
        """
        Gets dynamic axes that can be shared.

        You should try your best to give matching names to all axes that
        could potentially be shared (i.e. they have the same dimensions.)

        For example, if you know the second and third axes of multiple
        input tensors will match the latent height/width, you could do something like this:
            return {
                "tensor1": {2: "H", 3: "W"},
                "tensor2": {2: "H", 3: "W"}
            }
        The actual value given to the axis is unimportant, it needs only be consistent.
        """
        return {
            "sample": {0: "2B", 2: "H", 3: "W"},
            "encoder_hidden_states": {0: "2B"},
            "controlnet_cond": {0: "2B", 2: "HP", 3: "WP"},
            "mid_block": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "down_block_0": {0: "2B", 1: "S", 2: "H", 3: "W"},
            "down_block_1": {0: "2B", 1: "S", 2: "H", 3: "W"},
            "down_block_2": {0: "2B", 1: "S", 2: "H", 3: "W"},
            "down_block_3": {0: "2B", 1: "S", 2: "2H", 3: "2W"},
            "down_block_4": {0: "2B", 1: "2S", 2: "2H", 3: "2W"},
            "down_block_5": {0: "2B", 1: "2S", 2: "2H", 3: "2W"},
            "down_block_6": {0: "2B", 1: "2S", 2: "4H", 3: "4W"},
            "down_block_7": {0: "2B", 1: "4S", 2: "4H", 3: "4W"},
            "down_block_8": {0: "2B", 1: "4S", 2: "4H", 3: "4W"},
            "down_block_9": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "down_block_10": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "down_block_11": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
        }

    def get_input_profile(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Dict[str, List[Tuple[int, ...]]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        (
            (min_batch, max_batch),
            _,
            _,
            (min_latent_height, max_latent_height),
            (min_latent_width, max_latent_width),
        ) = self.get_minmax_dims(batch_size, image_height, image_width, static_batch, static_shape)
        return {
            "sample": [
                (2 * min_batch, self.unet_dim, min_latent_height, min_latent_width),
                (2 * batch_size, self.unet_dim, latent_height, latent_width),
                (2 * max_batch, self.unet_dim, max_latent_height, max_latent_width),
            ],
            "encoder_hidden_states": [
                (2 * min_batch, self.text_maxlen, self.embedding_dim),
                (2 * batch_size, self.text_maxlen, self.embedding_dim),
                (2 * max_batch, self.text_maxlen, self.embedding_dim),
            ],
            "controlnet_cond": [
                (2 * min_batch, 3, image_height, image_width),
                (2 * batch_size, 3, image_height, image_width),
                (2 * max_batch, 3, image_height, image_width),
            ],
        }

    def get_shape_dict(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Dict[str, Tuple[int, ...]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        return {
            "sample": (2 * batch_size, self.unet_dim, latent_height, latent_width),
            "encoder_hidden_states": (
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
            ),
            "controlnet_cond": (2 * batch_size, 3, image_height, image_width),
            "mid_block": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "down_block_0": (2 * batch_size, 320, latent_height, latent_width),
            "down_block_1": (2 * batch_size, 320, latent_height, latent_width),
            "down_block_2": (2 * batch_size, 320, latent_height, latent_width),
            "down_block_3": (2 * batch_size, 320, latent_height // 2, latent_width // 2),
            "down_block_4": (2 * batch_size, 640, latent_height // 2, latent_width // 2),
            "down_block_5": (2 * batch_size, 640, latent_height // 2, latent_width // 2),
            "down_block_6": (2 * batch_size, 640, latent_height // 4, latent_width // 4),
            "down_block_7": (2 * batch_size, 1280, latent_height // 4, latent_width // 4),
            "down_block_8": (2 * batch_size, 1280, latent_height // 4, latent_width // 4),
            "down_block_9": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "down_block_10": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "down_block_11": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
        }

    def get_sample_input(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
    ) -> Tuple[torch.Tensor, ...]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        dtype = torch.float16 if self.use_fp16 else torch.float32
        return (
            torch.randn(
                2 * batch_size,
                self.unet_dim,
                latent_height,
                latent_width,
                dtype=torch.float32,
                device=self.device,
            ),  # Sample
            torch.tensor([1.0], dtype=torch.float32, device=self.device),  # Timestep
            torch.randn(
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
                dtype=dtype,
                device=self.device,
            ),  # Encoder Hidden Sates
            torch.randn(
                2 * batch_size,
                3,
                image_height,
                image_width,
                dtype=torch.half,
                device=self.device,
            ),  # Controlnet Cond
            torch.randn(1, 1, 1, 1, dtype=torch.float32, device=self.device),  # Conditioning Scale
        )



from __future__ import annotations

from typing import Optional, Any, Dict, Iterator, Literal, Any

from html2text import HTML2Text
from pibble.api.client.webservice.jsonapi import JSONWebServiceAPIClient
from pibble.util.strings import Serializer

__all__ = ["CivitAIModel", "CivitAI"]


class CivitAIModel:
    """
    This represents an individual result from CivitAI
    """

    parser: HTML2Text

    @property
    def html_parser(self) -> HTML2Text:
        """
        Gets the parser that turns HTML to plaintext.
        """
        if not hasattr(CivitAIModel, "parser"):
            CivitAIModel.parser = HTML2Text()
            CivitAIModel.parser.ignore_links = True
            CivitAIModel.parser.ignore_emphasis = True
            CivitAIModel.parser.bypass_tables = True
        return CivitAIModel.parser

    def __init__(self, **properties: Any) -> None:
        """
        On initialization, clean description.
        """
        self.properties = properties
        if self.properties.get("description", None) is not None:
            lines = self.html_parser.handle(self.properties["description"]).splitlines()
            self.properties["description"] = "\n".join(
                [line.strip() for line in lines if line.strip()]
            )

    def serialize(self) -> Dict[str, Any]:
        """
        When serializing just return the properties dict.
        """
        return self.properties

    def __str__(self) -> str:
        """
        When stringifying, run through the serializer.
        """
        return Serializer.serialize(self.serialize())


class CivitAI(JSONWebServiceAPIClient):
    """
    This class provides some easy abstractions for querying from CivitAI.
    """

    PAGE_SIZE = 100

    def configure(self, client: Optional[Dict[str, Any]] = None, **kwargs: Any) -> None:
        if client is None:
            client = {}
        client["secure"] = True
        client["host"] = "civitai.com"
        client["path"] = "/api/v1"
        super(CivitAI, self).configure(client=client, **kwargs)

    def get_all(
        self, path: str, limit: Optional[int] = None, **parameters: Any
    ) -> Iterator[Dict[str, Any]]:
        """
        Gets all results from an endpoint, paginating if necessary
        """

        parameters = dict(
            [(key, value) for key, value in parameters.items() if value is not None and value != ""]
        )

        if limit is not None:
            parameters["limit"] = min(limit, self.PAGE_SIZE)

        # Get first page
        page = self.get(path, parameters=parameters).json()
        total_items = 0

        while page is not None:
            for item in page["items"]:
                total_items += 1
                yield item
                if limit is not None and total_items >= limit:
                    return
            if page["metadata"].get("nextPage", None) is not None:
                # Remove limit, it gets sent in the next page URL
                page = self.get(page["metadata"]["nextPage"]).json()
            else:
                page = None

    def get_models(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        query: Optional[str] = None,
        tag: Optional[str] = None,
        username: Optional[str] = None,
        types: Optional[
            Literal[
                "Checkpoint",
                "TextualInversion",
                "Hypernetwork",
                "AestheticGradient",
                "LORA",
                "Controlnet",
                "Poses",
            ]
        ] = None,
        sort: Optional[Literal["Highest Rated", "Most Downloaded", "Newest"]] = None,
        period: Optional[Literal["AllTime", "Year", "Month", "Week", "Day"]] = None,
        rating: Optional[int] = None,
        allowNoCredit: Optional[bool] = None,
        allowDerivatives: Optional[bool] = None,
        allowDifferentLicenses: Optional[bool] = None,
        allowCommercialUse: Optional[Literal["None", "Image", "Rent", "Sell"]] = None,
        nsfw: Optional[bool] = None,
    ) -> Iterator[CivitAIModel]:
        """
        Gets models, optionally filtered by some criteria.
        """
        for model in self.get_all(
            "/models",
            limit=limit,
            page=page,
            query=query,
            tag=tag,
            username=username,
            types=types,
            sort=sort,
            period=period,
            rating=rating,
            allowNoCredit=allowNoCredit,
            allowDerivatives=allowDerivatives,
            allowDifferentLicenses=allowDifferentLicenses,
            allowCommercialUse=allowCommercialUse,
            nsfw=nsfw,
        ):
            yield CivitAIModel(**model)



from __future__ import annotations

import os
import logging
import datetime
import traceback

import PIL
import PIL.Image

from typing import (
    Any,
    Optional,
    Callable,
    Union,
    List,
    Dict,
    Tuple,
    TYPE_CHECKING,
)

from multiprocessing import Process
from multiprocessing.queues import Queue
from queue import Empty

from pibble.api.configuration import APIConfiguration
from pibble.util.helpers import qualify
from pibble.util.strings import Serializer, get_uuid

from enfugue.util import logger

if TYPE_CHECKING:
    # We only import these here when type checking.
    # We avoid importing them before the process starts at runtime,
    # since we don't want torch to initialize itself.
    from enfugue.diffusion.manager import DiffusionPipelineManager
    from enfugue.diffusion.plan import DiffusionPlan

__all__ = ["DiffusionEngineProcess"]


class DiffusionEngineProcess(Process):
    """
    This process allows for easy two-way communication with a waiting
    Stable Diffusion Pipeline. Torch is only initiated after the process
    has began.
    """

    POLLING_DELAY_MS = 500
    IDLE_SEC = 30

    def __init__(
        self,
        instructions: Queue,
        results: Queue,
        intermediates: Queue,
        configuration: Optional[APIConfiguration] = None,
    ) -> None:
        super(DiffusionEngineProcess, self).__init__()
        self.configuration = APIConfiguration()
        self.instructions = instructions
        self.results = results
        self.intermediates = intermediates
        if configuration is not None:
            self.configuration = configuration

    @property
    def pipemanager(self) -> DiffusionPipelineManager:
        """
        Gets the model manager.
        """
        if not hasattr(self, "_pipemanager"):
            from enfugue.diffusion.manager import DiffusionPipelineManager

            self._pipemanager = DiffusionPipelineManager(self.configuration)
        return self._pipemanager

    @property
    def idle_seconds(self) -> int:
        """
        Gets the maximum number of seconds to go idle before exiting.
        """
        return self.configuration.get("enfugue.idle", self.IDLE_SEC)

    def get_diffusion_plan(self, payload: Dict[str, Any]) -> DiffusionPlan:
        """
        Deserializes a plan.
        """
        from enfugue.diffusion.plan import DiffusionPlan

        return DiffusionPlan.deserialize_dict(payload)

    def execute_diffusion_plan(
        self,
        instruction_id: int,
        plan: DiffusionPlan,
        intermediate_dir: Optional[str] = None,
        intermediate_steps: Optional[int] = None,
    ) -> List[PIL.Image.Image]:
        """
        Executes the plan, getting callbacks first.
        """

        progress_callback = self.create_progress_callback(instruction_id)
        if intermediate_dir is not None:
            image_callback = self.create_image_callback(
                instruction_id, intermediate_dir=intermediate_dir
            )
        else:
            image_callback = None

        self.pipemanager.keepalive_callback = lambda: progress_callback(0, 0, 0.0)

        return plan.execute(
            self.pipemanager,
            image_callback=image_callback,
            image_callback_steps=intermediate_steps,
            progress_callback=progress_callback,
        )

    def create_progress_callback(
        self,
        instruction_id: int,
    ) -> Callable[[int, int, float], None]:
        """
        Generates a callback that sends progress to the pipe.
        """

        def callback(current_step: int, total_steps: int, current_rate: float) -> None:
            to_send: Dict[str, Any] = {
                "id": instruction_id,
                "step": current_step,
                "total": total_steps,
                "rate": current_rate,
            }
            self.intermediates.put_nowait(Serializer.serialize(to_send))

        return callback

    def create_image_callback(
        self, instruction_id: int, intermediate_dir: str
    ) -> Callable[[List[PIL.Image.Image]], None]:
        """
        Generates a callback that sends decoded latents to the pipe, if asked.
        """

        def callback(images: List[PIL.Image.Image]) -> None:
            image_id = get_uuid()
            to_send: Dict[str, Any] = {"id": instruction_id, "images": []}
            for i, image in enumerate(images):
                image_path = os.path.join(intermediate_dir, f"{instruction_id}_{image_id}_{i}.png")
                image.save(image_path)
                to_send["images"].append(image_path)
            self.intermediates.put_nowait(Serializer.serialize(to_send))

        return callback

    def check_invoke_kwargs(
        self,
        instruction_id: int,
        model: Optional[str] = None,
        lora: Optional[Union[str, Tuple[str, float], List[Union[str, Tuple[str, float]]]]] = None,
        inversion: Optional[Union[str, List[str]]] = None,
        controlnet: Optional[str] = None,
        control_image: Optional[Union[str, PIL.Image.Image]] = None,
        seed: Optional[int] = None,
        image_callback_steps: Optional[int] = None,
        build_tensorrt: Optional[bool] = None,
        intermediate_dir: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        chunking_size: Optional[int] = None,
        guidance_scale: Optional[float] = None,
        num_inference_steps: Optional[int] = None,
        size: Optional[int] = None,
        **kwargs: Any,
    ) -> dict:
        """
        Sets local vars which will rebuild the pipeline if required
        """
        kwargs["progress_callback"] = self.create_progress_callback(instruction_id)
        if intermediate_dir is not None and image_callback_steps is not None:
            kwargs["latent_callback"] = self.create_image_callback(
                instruction_id, intermediate_dir=intermediate_dir
            )
            kwargs["latent_callback_steps"] = image_callback_steps
            kwargs["latent_callback_type"] = "pil"
        else:
            kwargs["latent_callback"] = None

        self.pipemanager.keepalive_callback = lambda: kwargs["progress_callback"](0, 0, 0.0)

        if model is not None:
            self.pipemanager.model = model

        if seed is not None:
            self.pipemanager.seed = seed

        if lora is not None:
            self.pipemanager.lora = lora  # type: ignore

        if inversion is not None:
            self.pipemanager.inversion = inversion  # type: ignore

        if build_tensorrt is not None:
            self.pipemanager.build_tensorrt = build_tensorrt

        if controlnet is not None:
            self.pipemanager.controlnet = controlnet  # type: ignore
            if control_image is not None:
                if isinstance(control_image, str):
                    control_image = PIL.Image.open(control_image)
                if controlnet == "canny":
                    control_image = self.pipemanager.edge_detector.canny(control_image)
                elif controlnet == "mlsd":
                    control_image = self.pipemanager.edge_detector.mlsd(control_image)
                elif controlnet == "hed":
                    control_image = self.pipemanager.edge_detector.hed(control_image)
                kwargs["control_image"] = control_image

        if size is not None:
            self.pipemanager.size = size

        if width is not None:
            kwargs["width"] = int(width)

        if height is not None:
            kwargs["height"] = int(height)

        if chunking_size is not None:
            kwargs["chunking_size"] = int(chunking_size)

        if guidance_scale is not None:
            kwargs["guidance_scale"] = float(guidance_scale)

        if num_inference_steps is not None:
            kwargs["num_inference_steps"] = int(num_inference_steps)

        return kwargs

    def clear_intermediates(self, instruction_id: int) -> None:
        """
        Clears intermediates for a specific instruction ID.
        """
        try:
            while True:
                next_intermediate = self.intermediates.get_nowait()
                # Avoid parsing
                if f'"id": {instruction_id}' not in next_intermediate[:40]:
                    # Not ours, put back on the queue
                    self.intermediates.put_nowait(next_intermediate)
        except Empty:
            return

    def clear_responses(self, instruction_id: int) -> None:
        """
        Clears responses for a specific instruction ID
        """
        try:
            while True:
                next_result = self.results.get_nowait()
                # Avoid parsing
                if f'"id": {instruction_id}' not in next_result[:40]:
                    # Not ours, put back on the queue
                    self.results.put_nowait(next_result)
        except Empty:
            return

    def run(self) -> None:
        """
        This is the function that the process will run.
        First instantiate the diffusion pipeline, then communicate as needed.
        """
        from pibble.util.helpers import OutputCatcher
        from pibble.util.log import ConfigurationLoggingContext
        with OutputCatcher():
            with ConfigurationLoggingContext(self.configuration):
                last_data = datetime.datetime.now()
                idle_seconds = 0.0

                while True:
                    try:
                        payload = self.instructions.get(timeout=self.POLLING_DELAY_MS / 1000)
                    except KeyboardInterrupt:
                        return
                    except Empty:
                        idle_seconds = (datetime.datetime.now() - last_data).total_seconds()
                        if idle_seconds > self.idle_seconds:
                            logger.info(
                                f"Reached maximum idle time after {idle_seconds:.1f} seconds, exiting engine process"
                            )
                            return
                        continue
                    except Exception as ex:
                        if logger.isEnabledFor(logging.DEBUG):
                            logger.debug(traceback.format_exc())
                        raise IOError(
                            "Received unexpected {0}, process will exit. {1}".format(
                                type(ex).__name__, ex
                            )
                        )

                    instruction = Serializer.deserialize(payload)
                    if not isinstance(instruction, dict):
                        logger.error(f"Unexpected non-dictionary argument {instruction}")
                        continue

                    instruction_id = instruction["id"]
                    instruction_action = instruction["action"]
                    instruction_payload = instruction.get("payload", None)

                    logger.debug(f"Received instruction {instruction_id}, action {instruction_action}")
                    if instruction_action == "ping":
                        logger.debug("Responding with 'pong'")
                        self.results.put(Serializer.serialize({"id": instruction_id, "result": "pong"}))
                    elif instruction_action in ["exit", "stop"]:
                        logger.debug("Exiting process")
                        self.pipemanager.unload_pipeline()
                        self.pipemanager.unload_upscaler()
                        return
                    elif instruction_action in ["invoke", "plan"]:
                        response = {"id": instruction_id, "payload": instruction_payload}
                        try:
                            if instruction_action == "plan":
                                intermediate_dir = instruction_payload.get("intermediate_dir", None)
                                intermediate_steps = instruction_payload.get("intermediate_steps", None)
                                plan = self.get_diffusion_plan(instruction_payload)
                                response["result"] = self.execute_diffusion_plan(
                                    instruction_id,
                                    plan,
                                    intermediate_dir=intermediate_dir,
                                    intermediate_steps=intermediate_steps,
                                )
                            else:
                                payload = self.check_invoke_kwargs(
                                    instruction_id, **instruction_payload
                                )
                                response["result"] = self.pipemanager(**payload)
                        except Exception as ex:
                            response["error"] = qualify(type(ex))
                            response["message"] = str(ex)
                            if logger.isEnabledFor(logging.DEBUG):
                                response["trace"] = traceback.format_exc()

                        del self.pipemanager.keepalive_callback
                        self.results.put(Serializer.serialize(response))
                        self.clear_intermediates(instruction_id)
                    else:
                        self.results.put(
                            Serializer.serialize(
                                {
                                    "id": instruction_id,
                                    "error": f"Unknown action '{instruction_action}'",
                                }
                            )
                        )
                    last_data = datetime.datetime.now()



import torch
from enfugue.diffusion.rt.model.base import BaseModel
from typing import Dict, List, Tuple


class UNet(BaseModel):
    def __init__(
        self,
        model: torch.nn.Module,
        use_fp16: bool = True,
        device: str = "cuda",
        max_batch_size: int = 16,
        embedding_dim: int = 768,
        text_maxlen: int = 77,
        unet_dim: int = 4,
    ) -> None:
        super(UNet, self).__init__(
            model=model,
            use_fp16=use_fp16,
            device=device,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
            text_maxlen=text_maxlen,
        )
        self.unet_dim = unet_dim
        self.name = "UNet"

    def get_model_key(self) -> str:
        return "unet"

    def get_input_names(self) -> List[str]:
        return ["sample", "timestep", "encoder_hidden_states"]

    def get_output_names(self) -> List[str]:
        return ["latent"]

    def get_dynamic_axes(self) -> Dict[str, Dict[int, str]]:
        return {
            "sample": {0: "2B", 2: "H", 3: "W"},
            "encoder_hidden_states": {0: "2B"},
            "latent": {0: "2B", 2: "H", 3: "W"},
        }

    def get_input_profile(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Dict[str, List[Tuple[int, ...]]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        (
            (min_batch, max_batch),
            _,
            _,
            (min_latent_height, max_latent_height),
            (min_latent_width, max_latent_width),
        ) = self.get_minmax_dims(batch_size, image_height, image_width, static_batch, static_shape)
        return {
            "sample": [
                (2 * min_batch, self.unet_dim, min_latent_height, min_latent_width),
                (2 * batch_size, self.unet_dim, latent_height, latent_width),
                (2 * max_batch, self.unet_dim, max_latent_height, max_latent_width),
            ],
            "encoder_hidden_states": [
                (2 * min_batch, self.text_maxlen, self.embedding_dim),
                (2 * batch_size, self.text_maxlen, self.embedding_dim),
                (2 * max_batch, self.text_maxlen, self.embedding_dim),
            ],
        }

    def get_shape_dict(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Dict[str, Tuple[int, ...]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        return {
            "sample": (2 * batch_size, self.unet_dim, latent_height, latent_width),
            "encoder_hidden_states": (
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
            ),
            "latent": (2 * batch_size, 4, latent_height, latent_width),
        }

    def get_sample_input(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
    ) -> Tuple[torch.Tensor, ...]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        dtype = torch.float16 if self.use_fp16 else torch.float32
        return (
            torch.randn(
                2 * batch_size,
                self.unet_dim,
                latent_height,
                latent_width,
                dtype=torch.float32,
                device=self.device,
            ),
            torch.tensor([1.0], dtype=torch.float32, device=self.device),
            torch.randn(
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
                dtype=dtype,
                device=self.device,
            ),
        )



import torch
from enfugue.diffusion.rt.model.base import BaseModel
from typing import Dict, List, Tuple


class ControlledUNet(BaseModel):
    """
    The controlled UNet uses a controlnet that forwards mid/down blocks.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        use_fp16: bool = True,
        device: str = "cuda",
        max_batch_size: int = 16,
        embedding_dim: int = 768,
        text_maxlen: int = 77,
        unet_dim: int = 4,
    ) -> None:
        super(ControlledUNet, self).__init__(
            model=model,
            use_fp16=use_fp16,
            device=device,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
            text_maxlen=text_maxlen,
        )
        self.unet_dim = unet_dim
        self.name = "ControlledUNet"

    def get_model_key(self) -> str:
        return "controlledunet"

    def get_input_names(self) -> List[str]:
        return [
            "sample",
            "timestep",
            "encoder_hidden_states",
            "mid_block",
            "down_block_0",
            "down_block_1",
            "down_block_2",
            "down_block_3",
            "down_block_4",
            "down_block_5",
            "down_block_6",
            "down_block_7",
            "down_block_8",
            "down_block_9",
            "down_block_10",
            "down_block_11",
        ]

    def get_output_names(self) -> List[str]:
        return ["latent"]

    def get_dynamic_axes(self) -> Dict[str, Dict[int, str]]:
        return {
            "sample": {0: "2B", 2: "H", 3: "W"},
            "encoder_hidden_states": {0: "2B"},
            "mid_block": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "down_block_0": {0: "2B", 1: "S", 2: "H", 3: "W"},
            "down_block_1": {0: "2B", 1: "S", 2: "H", 3: "W"},
            "down_block_2": {0: "2B", 1: "S", 2: "H", 3: "W"},
            "down_block_3": {0: "2B", 1: "S", 2: "2H", 3: "2W"},
            "down_block_4": {0: "2B", 1: "2S", 2: "2H", 3: "2W"},
            "down_block_5": {0: "2B", 1: "2S", 2: "2H", 3: "2W"},
            "down_block_6": {0: "2B", 1: "2S", 2: "4H", 3: "4W"},
            "down_block_7": {0: "2B", 1: "4S", 2: "4H", 3: "4W"},
            "down_block_8": {0: "2B", 1: "4S", 2: "4H", 3: "4W"},
            "down_block_9": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "down_block_10": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "down_block_11": {0: "2B", 1: "4S", 2: "8H", 3: "8W"},
            "latent": {0: "2B", 2: "H", 3: "W"},
        }

    def get_input_profile(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Dict[str, List[Tuple[int, ...]]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        (
            (min_batch, max_batch),
            _,
            _,
            (min_latent_height, max_latent_height),
            (min_latent_width, max_latent_width),
        ) = self.get_minmax_dims(batch_size, image_height, image_width, static_batch, static_shape)
        return {
            "sample": [
                (2 * min_batch, self.unet_dim, min_latent_height, min_latent_width),
                (2 * batch_size, self.unet_dim, latent_height, latent_width),
                (2 * max_batch, self.unet_dim, max_latent_height, max_latent_width),
            ],
            "encoder_hidden_states": [
                (2 * min_batch, self.text_maxlen, self.embedding_dim),
                (2 * batch_size, self.text_maxlen, self.embedding_dim),
                (2 * max_batch, self.text_maxlen, self.embedding_dim),
            ],
            "mid_block": [
                (2 * min_batch, 1280, min_latent_height // 8, min_latent_width // 8),
                (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
                (2 * max_batch, 1280, max_latent_height // 8, max_latent_width // 8),
            ],
            "down_block_0": [
                (2 * min_batch, 320, min_latent_height, min_latent_width),
                (2 * batch_size, 320, latent_height, latent_width),
                (2 * max_batch, 320, max_latent_height, max_latent_width),
            ],
            "down_block_1": [
                (2 * min_batch, 320, min_latent_height, min_latent_width),
                (2 * batch_size, 320, latent_height, latent_width),
                (2 * max_batch, 320, max_latent_height, max_latent_width),
            ],
            "down_block_2": [
                (2 * min_batch, 320, min_latent_height, min_latent_width),
                (2 * batch_size, 320, latent_height, latent_width),
                (2 * max_batch, 320, max_latent_height, max_latent_width),
            ],
            "down_block_3": [
                (2 * min_batch, 320, min_latent_height // 2, min_latent_width // 2),
                (2 * batch_size, 320, latent_height // 2, latent_width // 2),
                (2 * max_batch, 320, max_latent_height // 2, max_latent_width // 2),
            ],
            "down_block_4": [
                (2 * min_batch, 640, min_latent_height // 2, min_latent_width // 2),
                (2 * batch_size, 640, latent_height // 2, latent_width // 2),
                (2 * max_batch, 640, max_latent_height // 2, max_latent_width // 2),
            ],
            "down_block_5": [
                (2 * min_batch, 640, min_latent_height // 2, min_latent_width // 2),
                (2 * batch_size, 640, latent_height // 2, latent_width // 2),
                (2 * max_batch, 640, max_latent_height // 2, max_latent_width // 2),
            ],
            "down_block_6": [
                (2 * min_batch, 640, min_latent_height // 4, min_latent_width // 4),
                (2 * batch_size, 640, latent_height // 4, latent_width // 4),
                (2 * max_batch, 640, max_latent_height // 4, max_latent_width // 4),
            ],
            "down_block_7": [
                (2 * min_batch, 1280, min_latent_height // 4, min_latent_width // 4),
                (2 * batch_size, 1280, latent_height // 4, latent_width // 4),
                (2 * max_batch, 1280, max_latent_height // 4, max_latent_width // 4),
            ],
            "down_block_8": [
                (2 * min_batch, 1280, min_latent_height // 4, min_latent_width // 4),
                (2 * batch_size, 1280, latent_height // 4, latent_width // 4),
                (2 * max_batch, 1280, max_latent_height // 4, max_latent_width // 4),
            ],
            "down_block_9": [
                (2 * min_batch, 1280, min_latent_height // 8, min_latent_width // 8),
                (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
                (2 * max_batch, 1280, max_latent_height // 8, max_latent_width // 8),
            ],
            "down_block_10": [
                (2 * min_batch, 1280, min_latent_height // 8, min_latent_width // 8),
                (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
                (2 * max_batch, 1280, max_latent_height // 8, max_latent_width // 8),
            ],
            "down_block_11": [
                (2 * min_batch, 1280, min_latent_height // 8, min_latent_width // 8),
                (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
                (2 * max_batch, 1280, max_latent_height // 8, max_latent_width // 8),
            ],
        }

    def get_shape_dict(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Dict[str, Tuple[int, ...]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        return {
            "sample": (2 * batch_size, self.unet_dim, latent_height, latent_width),
            "encoder_hidden_states": (
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
            ),
            "mid_block": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "down_block_0": (2 * batch_size, 320, latent_height, latent_width),
            "down_block_1": (2 * batch_size, 320, latent_height, latent_width),
            "down_block_2": (2 * batch_size, 320, latent_height, latent_width),
            "down_block_3": (2 * batch_size, 320, latent_height // 2, latent_width // 2),
            "down_block_4": (2 * batch_size, 640, latent_height // 2, latent_width // 2),
            "down_block_5": (2 * batch_size, 640, latent_height // 2, latent_width // 2),
            "down_block_6": (2 * batch_size, 640, latent_height // 4, latent_width // 4),
            "down_block_7": (2 * batch_size, 1280, latent_height // 4, latent_width // 4),
            "down_block_8": (2 * batch_size, 1280, latent_height // 4, latent_width // 4),
            "down_block_9": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "down_block_10": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "down_block_11": (2 * batch_size, 1280, latent_height // 8, latent_width // 8),
            "latent": (2 * batch_size, 4, latent_height, latent_width),
        }

    def get_sample_input(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
    ) -> Tuple[torch.Tensor, ...]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        dtype = torch.float16 if self.use_fp16 else torch.float32
        return (
            torch.randn(
                2 * batch_size,
                self.unet_dim,
                latent_height,
                latent_width,
                dtype=torch.float32,
                device=self.device,
            ),  # Sample
            torch.tensor([1.0], dtype=torch.float32, device=self.device),  # Timestep
            torch.randn(
                2 * batch_size,
                self.text_maxlen,
                self.embedding_dim,
                dtype=dtype,
                device=self.device,
            ),  # Encoder Hidden Sates
            torch.randn(
                2 * batch_size,
                1280,
                latent_height // 8,
                latent_width // 8,
                dtype=dtype,
                device=self.device,
            ),  # mid block
            torch.randn(
                2 * batch_size,
                320,
                latent_height,
                latent_width,
                dtype=dtype,
                device=self.device,
            ),  # down block 0
            torch.randn(
                2 * batch_size,
                320,
                latent_height,
                latent_width,
                dtype=dtype,
                device=self.device,
            ),  # down block 1
            torch.randn(
                2 * batch_size,
                320,
                latent_height,
                latent_width,
                dtype=dtype,
                device=self.device,
            ),  # down block 2
            torch.randn(
                2 * batch_size,
                320,
                latent_height // 2,
                latent_width // 2,
                dtype=dtype,
                device=self.device,
            ),  # down block 3
            torch.randn(
                2 * batch_size,
                640,
                latent_height // 2,
                latent_width // 2,
                dtype=dtype,
                device=self.device,
            ),  # down block 4
            torch.randn(
                2 * batch_size,
                640,
                latent_height // 2,
                latent_width // 2,
                dtype=dtype,
                device=self.device,
            ),  # down block 5
            torch.randn(
                2 * batch_size,
                640,
                latent_height // 4,
                latent_width // 4,
                dtype=dtype,
                device=self.device,
            ),  # down block 6
            torch.randn(
                2 * batch_size,
                1280,
                latent_height // 4,
                latent_width // 4,
                dtype=dtype,
                device=self.device,
            ),  # down block 7
            torch.randn(
                2 * batch_size,
                1280,
                latent_height // 4,
                latent_width // 4,
                dtype=dtype,
                device=self.device,
            ),  # down block 8
            torch.randn(
                2 * batch_size,
                1280,
                latent_height // 8,
                latent_width // 8,
                dtype=dtype,
                device=self.device,
            ),  # down block 9
            torch.randn(
                2 * batch_size,
                1280,
                latent_height // 8,
                latent_width // 8,
                dtype=dtype,
                device=self.device,
            ),  # down block 10
            torch.randn(
                2 * batch_size,
                1280,
                latent_height // 8,
                latent_width // 8,
                dtype=dtype,
                device=self.device,
            ),  # down block 11
        )



import os
import shutil

from typing import Dict, Any, Tuple, List

from sqlalchemy import delete

from webob import Request, Response

from pibble.api.server.webservice.jsonapi import JSONWebServiceAPIServer
from pibble.api.exceptions import BadRequestError, NotFoundError, ConfigurationError
from pibble.ext.user.database import User
from pibble.ext.user.server.base import UserExtensionHandlerRegistry
from pibble.util.encryption import Password

from enfugue.api.controller.base import EnfugueAPIControllerBase

__all__ = ["EnfugueAPISystemController"]


def get_directory_size(directory: str, recurse: bool = True) -> Tuple[int, int, int]:
    """
    Sums the files and filesize of a directory
    """
    if not os.path.exists(directory):
        return 0, 0, 0
    items = os.listdir(directory)
    top_level_items = len(items)
    files, size = 0, 0
    for item in items:
        path = os.path.join(directory, item)
        if os.path.isfile(path):
            files += 1
            size += os.path.getsize(path)
        elif recurse:
            sub_items, sub_files, sub_size = get_directory_size(path, recurse=True)
            files += sub_files
            size += sub_size
    return top_level_items, files, size


class EnfugueAPISystemController(EnfugueAPIControllerBase):
    handlers = UserExtensionHandlerRegistry()

    @handlers.path("^/api/settings$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("System", "read")
    def get_settings(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        Gets the settings that can be manipulated from the UI
        """
        return {
            "safe": self.configuration.get("enfugue.safe", True),
            "auth": not (self.configuration.get("enfugue.noauth", True)),
            "max_queued_invocations": self.manager.max_queued_invocations,
            "max_queued_downloads": self.manager.max_queued_downloads,
            "max_concurrent_downloads": self.manager.max_concurrent_downloads,
        }

    @handlers.path("^/api/settings$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("System", "update")
    def update_settings(self, request: Request, response: Response) -> None:
        """
        Updates the settings that can be manipulated from the UI
        """
        if "auth" in request.parsed:
            self.user_config["enfugue.noauth"] = not request.parsed["auth"]
            if self.user_config["enfugue.noauth"] != request.parsed["auth"]:
                self.database.execute(delete(self.orm.AuthenticationToken))  # Clear auth data
                if request.parsed["auth"]:
                    self.database.execute(
                        delete(self.orm.User).filter(self.orm.User.username == "noauth")
                    )
                self.database.commit()
        if "safe" in request.parsed:
            if request.parsed["safe"] != self.configuration.get("enfugue.safe", True):
                # Destroy process so we re-initialize with new safety settings
                self.manager.stop_engine()
        for key in [
            "safe",
            "max_queued_invocation",
            "max_queued_downloads",
            "max_concurrent_downloads",
        ]:
            if key in request.parsed:
                self.user_config[f"enfugue.{key}"] = request.parsed[key]
        self.configuration.update(**self.user_config.dict())

    @handlers.path("^/api/users$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("User", "create")
    def create_user(self, request: Request, response: Response) -> User:
        """
        Creates a user.
        """
        username = request.parsed.get("username", None)
        if not username:
            raise BadRequestError("Username is required.")
        user = (
            self.database.query(self.orm.User)
            .filter(self.orm.User.username == username)
            .one_or_none()
        )
        if user:
            raise BadRequestError(f"User {username} already exists.")
        password = request.parsed.get("new_password", None)
        repeat_password = request.parsed.get("repeat_password", None)

        if not password or not repeat_password:
            raise BadRequestError("Password is required.")
        if password != repeat_password:
            raise BadRequestError("Passwords do not match.")

        user = self.orm.User(username=username, password=Password.hash(password))

        if "first_name" in request.parsed:
            user.first_name = request.parsed["first_name"]
        if "last_name" in request.parsed:
            user.last_name = request.parsed["last_name"]

        self.database.add(user)
        self.database.commit()

        if "admin" in request.parsed and request.parsed["admin"]:
            admin_permission_group = (
                self.database.query(self.orm.PermissionGroup)
                .filter(self.orm.PermissionGroup.label == "admin")
                .one_or_none()
            )
            if not admin_permission_group:
                raise ConfigurationError(
                    "Couldn't find admin permission group. Did you modify the user initialization configuration?"
                )
            self.database.add(
                self.orm.UserPermissionGroup(user_id=user.id, group_id=admin_permission_group.id)
            )
        self.database.commit()
        return user

    @handlers.path("^/api/users/(?P<username>[a-zA-Z0-9_]+)$")
    @handlers.methods("PATCH")
    @handlers.format()
    @handlers.secured("User", "update")
    def update_user(self, request: Request, response: Response, username: str) -> User:
        """
        Updates one user.
        """
        user = (
            self.database.query(self.orm.User)
            .filter(self.orm.User.username == username)
            .one_or_none()
        )
        if not user:
            raise NotFoundError(f"No user named {username}")

        if "first_name" in request.parsed:
            user.first_name = request.parsed["first_name"]
        if "last_name" in request.parsed:
            user.last_name = request.parsed["last_name"]
        if "new_password" in request.parsed and "repeat_password" in request.parsed:
            if request.parsed["new_password"] != request.parsed["repeat_password"]:
                raise BadRequestError("Passwords do not match.")
            user.password = Password.hash(request.parsed["new_password"])
        if "admin" in request.parsed:
            if username == "enfugue" and not request.parsed["admin"]:
                raise BadRequestError("Cannot demote default user.")

            admin_permission_group = (
                self.database.query(self.orm.PermissionGroup)
                .filter(self.orm.PermissionGroup.label == "admin")
                .one_or_none()
            )
            if not admin_permission_group:
                raise ConfigurationError(
                    "Couldn't find admin permission group. Did you modify the user initialization configuration?"
                )
            admin_permission = None
            for user_permission_group in user.permission_groups:
                if user_permission_group.group_id == admin_permission_group.id:
                    admin_permission = user_permission_group
                    break

            if admin_permission is not None and not request.parsed["admin"]:
                self.database.delete(admin_permission)
            elif admin_permission is None and request.parsed["admin"]:
                self.database.add(
                    self.orm.UserPermissionGroup(
                        user_id=user.id, group_id=admin_permission_group.id
                    )
                )
        self.database.commit()
        return user

    @handlers.path("^/api/users/(?P<username>[a-zA-Z0-9_]+)$")
    @handlers.methods("DELETE")
    @handlers.format()
    @handlers.secured("User", "delete")
    def delete_user(self, request: Request, response: Response, username: str) -> None:
        """
        Deletes one user.
        We have to do the cascading ourselves because of a bug with sqlite and sqlalchemy.
        """
        if username == "enfugue":
            raise BadRequestError("Cannot delete default user.")
        user = (
            self.database.query(self.orm.User)
            .filter(self.orm.User.username == username)
            .one_or_none()
        )
        if not user:
            raise NotFoundError(f"No user named {username}")
        for permission in user.permissions:
            self.database.delete(permission)
        for permission_group in user.permission_groups:
            self.database.delete(permission_group)
        self.database.commit()
        self.database.delete(user)
        self.database.commit()

    @handlers.path("^/api/installation$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("System", "read")
    def get_installation_summary(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        Gets a summary of files and filesize in the installation
        """
        sizes = {}
        for dirname in ["cache", "checkpoint", "inversion", "lora", "models", "other"]:
            items, files, size = get_directory_size(os.path.join(self.engine_root, dirname))
            sizes[dirname] = {"items": items, "files": files, "bytes": size}
        return sizes

    @handlers.path("^/api/installation/(?P<dirname>[a-zA-Z0-9_]+)$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("System", "read")
    def get_installation_details(
        self, request: Request, response: Response, dirname: str
    ) -> List[Dict[str, Any]]:
        """
        Gets a summary of files and filesize in the installation
        """
        directory = os.path.join(self.engine_root, dirname)
        if not os.path.isdir(directory):
            return []

        items = []

        for item in os.listdir(directory):
            path = os.path.join(directory, item)
            if os.path.isdir(path):
                sub_items, files, size = get_directory_size(path)
                items.append({"type": "directory", "name": item, "bytes": size})
            else:
                items.append({"type": "file", "name": item, "bytes": os.path.getsize(path)})
        return items

    @handlers.path("^/api/installation/(?P<dirname>[^\/]+)/(?P<filename>[^\/]+)$")
    @handlers.methods("DELETE")
    @handlers.format()
    @handlers.secured("System", "update")
    def remove_from_installation(
        self, request: Request, response: Response, dirname: str, filename: str
    ) -> None:
        """
        Deletes a file or directory from the installation
        """

        path = os.path.join(self.engine_root, dirname, filename)
        if not os.path.exists(path):
            raise BadRequestError(f"Unknown engine file/directory {dirname}/{filename}")
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)

    @handlers.bypass(JSONWebServiceAPIServer)
    @handlers.path("^/api/installation/(?P<dirname>[^\/]+)$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("System", "update")
    def add_to_installation(self, request: Request, response: Response, dirname: str) -> None:
        """
        Uploads a file to an installation directory.
        """

        if "file" not in request.POST:
            raise BadRequestError("File is missing.")

        filename = request.POST["file"].filename
        dirpath = os.path.join(self.engine_root, dirname)
        if not os.path.exists(dirpath):
            raise BadRequestError(f"Unknonwn directory {dirname}")

        path = os.path.join(dirpath, filename)
        with open(path, "wb") as handle:
            for chunk in request.POST["file"].file:
                handle.write(chunk)
